import logging
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Sequence, Union, cast

import numpy as np
import numpy.typing as npt
import sympy as smp

from centrex_tlf import constants, hamiltonian, states
from centrex_tlf import couplings as couplings_tlf
from centrex_tlf.couplings.utils_compact import (
    compact_coupling_field,
    insert_levels_coupling_field,
)
from centrex_tlf.transitions import MicrowaveTransition, OpticalTransition

from . import utils_decay as utils_decay
from .generate_hamiltonian import generate_total_symbolic_hamiltonian
from .generate_system_of_equations import generate_system_of_equations_symbolic
from .utils_compact import generate_qn_compact

__all__ = [
    "generate_OBE_system",
    "generate_OBE_system_transitions",
    "setup_OBE_system",
    "setup_OBE_system_transitions",
    "OBESystem",
]


@dataclass
class OBESystem:
    ground: Sequence[states.CoupledState]
    excited: Sequence[states.CoupledState]
    QN: Sequence[states.CoupledState]
    H_int: npt.NDArray[np.complex128]
    V_ref_int: npt.NDArray[np.complex128]
    couplings: List[Any]
    H_symbolic: smp.MutableDenseMatrix
    C_array: npt.NDArray[np.floating]
    system: smp.MutableDenseMatrix
    coupling_symbols: Sequence[smp.Symbol]
    polarization_symbols: Sequence[Sequence[smp.Symbol]]
    dissipator: Optional[smp.MutableDenseMatrix] = None
    QN_original: Optional[Sequence[states.CoupledState]] = None
    decay_channels: Optional[Sequence[utils_decay.DecayChannel]] = None
    couplings_original: Optional[List[couplings_tlf.CouplingFields]] = None

    def __repr__(self) -> str:
        ground = [s.largest for s in self.ground]
        ground = list(
            np.unique(
                [
                    f"|{s.electronic_state.name}, J = {s.J}, "  # type: ignore
                    f"P = {'+' if s.P == 1 else '-'}>"  # type: ignore
                    for s in ground
                ]
            )
        )
        ground_str: str = ", ".join(ground)  # type: ignore
        excited = [s.largest for s in self.excited]
        excited = list(
            np.unique(
                [
                    str(
                        f"|{s.electronic_state.name}, J = {s.J}, "  # type: ignore
                        f"F₁ = {smp.S(str(s.F1), rational=True)}, "  # type: ignore
                        f"F = {s.F}, "  # type: ignore
                        f"P = {'+' if s.P == 1 else '-'}>"  # type: ignore
                    )
                    for s in excited
                ]
            )
        )
        excited_str: str = ", ".join(excited)  # type: ignore
        return f"OBESystem(ground=[{ground_str}], excited=[{excited_str}])"


def check_transitions_allowed(
    transition_selectors: Sequence[couplings_tlf.TransitionSelector],
) -> None:
    """
    Check if a sequence of TransitionSelectors are all allowed transitions

    Args:
        transition_selectors (Sequence[couplings_TlF.TransitionSelector]): Sequence of
        TransitionSelectors a set of transitions.

    Raises:
        AssertionError: error if any given transition is not allowed.
    """
    for transition_selector in transition_selectors:
        if (
            transition_selector.ground_main is not None
            and transition_selector.excited_main is not None
        ):
            try:
                ΔmF_allowed = couplings_tlf.utils.ΔmF_allowed(
                    transition_selector.polarizations[0]
                )
                couplings_tlf.utils.assert_transition_coupled_allowed(
                    cast(
                        states.CoupledBasisState,
                        transition_selector.ground_main.largest,
                    ),
                    cast(
                        states.CoupledBasisState,
                        transition_selector.excited_main.largest,
                    ),
                    ΔmF_allowed=ΔmF_allowed,
                )
            except AssertionError as err:
                raise AssertionError(
                    f"{transition_selector.description} with polarization "
                    f"{np.round(transition_selector.polarizations[0], 2)} => "
                    f"{err.args[0]}"
                )
        else:
            raise ValueError(
                "Cannot check if transition is allowed; no main states selected"
            )
    return


def generate_OBE_system(
    X_states: Union[states.QuantumSelector, Sequence[states.QuantumSelector]],
    B_states: Union[states.QuantumSelector, Sequence[states.QuantumSelector]],
    transition_selectors: Sequence[couplings_tlf.TransitionSelector],
    qn_compact: Optional[
        Union[states.QuantumSelector, Sequence[states.QuantumSelector]]
    ] = None,
    decay_channels: Optional[
        Union[Sequence[utils_decay.DecayChannel], utils_decay.DecayChannel]
    ] = None,
    E: npt.NDArray[np.floating] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.floating] = np.array([0.0, 0.0, 1e-5]),
    X_constants: constants.XConstants = constants.XConstants(),
    B_constants: constants.BConstants = constants.BConstants(),
    nuclear_spins: constants.TlFNuclearSpins = constants.TlFNuclearSpins(),
    Jmin_X: Optional[int] = None,
    Jmax_X: Optional[int] = None,
    Jmin_B: Optional[int] = None,
    Jmax_B: Optional[int] = None,
    transform: Optional[npt.NDArray[np.complex128]] = None,
    H_func_X: Optional[Callable] = None,
    H_func_B: Optional[Callable] = None,
    verbose: bool = False,
    normalize_pol: bool = False,
    Γ: float = hamiltonian.Γ,
) -> OBESystem:
    """Convenience function for generating the symbolic OBE system of equations.

    Args:

        X_states : X states to include in the OBE system
        B_states : B states to include in the OBE system
        transition_selectors (list): list of TransitionSelectors defining the
                                    transitions used in the OBE system.
        qn_compact (QuantumSelector): dataclass specifying a subset of states to
                                        select based on the quantum numbers
        decay_channels (DecayChannel): dataclass specifying the decay channel to
                                        add
        verbose (bool, optional): Log progress to INFO. Defaults to False.

    Returns:
        OBESystem: dataclass designed to hold the generated values
                    ground, exxcited, QN, H_int, V_ref_int, couplings, H_symbolic,
                    C_array, system
    """
    # check if transitions are allowed before generating the hamiltonian
    check_transitions_allowed(transition_selectors=transition_selectors)

    QN_X_original = list(states.generate_coupled_states_X(X_states))
    QN_B_original = list(states.generate_coupled_states_B(B_states))
    QN_original = QN_X_original + QN_B_original
    rtol = None
    stol = 1e-3
    if verbose:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
        logger.info("generate_OBE_system: 1/5 -> Generating the reduced Hamiltonian")
    H_reduced = hamiltonian.generate_total_reduced_hamiltonian(
        X_states_approx=QN_X_original,
        B_states_approx=QN_B_original,
        E=E,
        B=B,
        Jmin_X=Jmin_X,
        Jmax_X=Jmax_X,
        Jmin_B=Jmin_B,
        Jmax_B=Jmax_B,
        rtol=rtol,
        stol=stol,
        X_constants=X_constants,
        B_constants=B_constants,
        nuclear_spins=nuclear_spins,
        transform=transform,
        H_func_X=H_func_X,
        H_func_B=H_func_B,
    )
    ground_states = H_reduced.X_states
    excited_states = H_reduced.B_states
    QN = H_reduced.QN
    H_int = H_reduced.H_int

    V_ref_int = H_reduced.V_ref_int
    if verbose:
        logger.info(
            "generate_OBE_system: 2/5 -> "
            "Generating the couplings corresponding to the transitions"
        )
    couplings = []
    for transition in transition_selectors:
        if transition.ground_main is not None and transition.excited_main is not None:
            couplings.append(
                couplings_tlf.generate_coupling_field(
                    transition.ground_main,
                    transition.excited_main,
                    transition.ground,
                    transition.excited,
                    QN_original,
                    H_int,
                    QN,
                    V_ref_int,
                    pol_vecs=transition.polarizations,
                    pol_main=transition.polarizations[0],
                    normalize_pol=normalize_pol,
                )
            )
        else:
            couplings.append(
                couplings_tlf.generate_coupling_field_automatic(
                    transition.ground,
                    transition.excited,
                    QN_original,
                    H_int,
                    QN,
                    V_ref_int,
                    pol_vecs=transition.polarizations,
                )
            )

    if verbose:
        logger.info("generate_OBE_system: 3/5 -> Generating the symbolic Hamiltonian")
    if qn_compact is not None:
        H_symbolic, QN_compact = generate_total_symbolic_hamiltonian(
            QN, H_int, couplings, transition_selectors, qn_compact=qn_compact
        )
        couplings_compact = [
            compact_coupling_field(coupling, QN, qn_compact) for coupling in couplings
        ]
    else:
        H_symbolic = generate_total_symbolic_hamiltonian(
            QN, H_int, couplings, transition_selectors
        )

    if verbose:
        logger.info("generate_OBE_system: 4/5 -> Generating the collapse matrices")
    C_array = couplings_tlf.collapse_matrices(
        QN,
        ground_states,
        excited_states,
        gamma=Γ,
        qn_compact=qn_compact,
    )

    if decay_channels is not None:
        if isinstance(decay_channels, list):
            _decay_channels = decay_channels
        elif not isinstance(decay_channels, (tuple, list, np.ndarray)):
            _decay_channels = [decay_channels]
        elif isinstance(decay_channels, (tuple, np.ndarray)):
            _decay_channels = list(decay_channels)
        else:
            raise AssertionError(
                f"decay_channels is type f{type(decay_channels)}; supply a list, tuple"
                " or np.ndarray"
            )

        indices = utils_decay.get_insert_level_indices(
            _decay_channels, QN, excited_states
        )
        couplings = [
            insert_levels_coupling_field(coupling, indices_insert=indices)
            for coupling in couplings
        ]
        QN = utils_decay.add_states_QN(_decay_channels, QN, indices)

        if qn_compact is not None:
            indices, H_symbolic = utils_decay.add_levels_symbolic_hamiltonian(
                H_symbolic, _decay_channels, QN_compact, excited_states
            )
            QN_compact = utils_decay.add_states_QN(_decay_channels, QN_compact, indices)
            C_array = utils_decay.add_decays_C_arrays(
                _decay_channels, indices, QN_compact, C_array, Γ
            )

            couplings_compact = [
                insert_levels_coupling_field(coupling, indices_insert=indices)
                for coupling in couplings_compact
            ]
        else:
            indices, H_symbolic = utils_decay.add_levels_symbolic_hamiltonian(
                H_symbolic, _decay_channels, QN, excited_states
            )
            C_array = utils_decay.add_decays_C_arrays(
                _decay_channels, indices, QN, C_array, Γ
            )

    if verbose:
        logger.info(
            "generate_OBE_system: 5/5 -> Transforming the Hamiltonian and collapse "
            "matrices into a symbolic system of equations"
        )
    system = generate_system_of_equations_symbolic(H_symbolic, C_array, fast=True)
    obe_system = OBESystem(
        QN=QN_compact if qn_compact is not None else QN,
        ground=ground_states,
        excited=excited_states,
        couplings=couplings if qn_compact is None else couplings_compact,
        H_symbolic=H_symbolic,
        H_int=H_int,
        V_ref_int=V_ref_int,
        C_array=C_array,
        system=system,
        coupling_symbols=[trans.Ω for trans in transition_selectors],
        polarization_symbols=[
            trans.polarization_symbols for trans in transition_selectors
        ],
        QN_original=None if qn_compact is None else QN,
        decay_channels=_decay_channels if decay_channels else None,
        couplings_original=None if qn_compact is None else couplings,
    )
    return obe_system


def generate_OBE_system_transitions(
    transitions: Sequence[Union[OpticalTransition, MicrowaveTransition]],
    transition_selectors: Sequence[couplings_tlf.TransitionSelector],
    qn_compact: Optional[
        Union[states.QuantumSelector, Sequence[states.QuantumSelector], bool]
    ] = None,
    decay_channels: Optional[
        Union[Sequence[utils_decay.DecayChannel], utils_decay.DecayChannel]
    ] = None,
    E: npt.NDArray[np.floating] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.floating] = np.array([0.0, 0.0, 1e-5]),
    Γ: float = hamiltonian.Γ,
    X_constants: constants.XConstants = constants.XConstants(),
    B_constants: constants.BConstants = constants.BConstants(),
    nuclear_spins: constants.TlFNuclearSpins = constants.TlFNuclearSpins(),
    Jmin_X: Optional[int] = None,
    Jmax_X: Optional[int] = None,
    Jmin_B: Optional[int] = None,
    Jmax_B: Optional[int] = None,
    transform: Optional[npt.NDArray[np.complex128]] = None,
    H_func_X: Optional[Callable] = None,
    H_func_B: Optional[Callable] = None,
    verbose: bool = False,
    normalize_pol: bool = False,
) -> OBESystem:
    """Convenience function for generating the symbolic OBE system of equations.

    Args:
        transitions (list): list of TransitionSelectors defining the transitions
                            used in the OBE system.
        qn_compact (QuantumSelector): dataclass specifying a subset of states to
                                        select based on the quantum numbers
        decay_channels (DecayChannel): dataclass specifying the decay channel to
                                        add
        verbose (bool, optional): Log progress to INFO. Defaults to False.

    Returns:
        OBESystem: dataclass designed to hold the generated values
                    ground, exxcited, QN, H_int, V_ref_int, couplings, H_symbolic,
                    C_array, system
    """
    rtol = None
    stol = 1e-3

    # check if transitions are allowed before generating the hamiltonian
    check_transitions_allowed(transition_selectors=transition_selectors)

    if verbose:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
        logger.info("generate_OBE_system: 1/5 -> Generating the reduced Hamiltonian")
    H_reduced = hamiltonian.generate_reduced_hamiltonian_transitions(
        transitions=transitions,
        E=E,
        B=B,
        rtol=rtol,
        stol=stol,
        Jmin_X=Jmin_X,
        Jmax_X=Jmax_X,
        Jmin_B=Jmin_B,
        Jmax_B=Jmax_B,
        Xconstants=X_constants,
        Bconstants=B_constants,
        nuclear_spins=nuclear_spins,
        # transform=transform,
        # H_func_X=H_func_X,
        # H_func_B=H_func_B,
    )

    if H_reduced.QN_basis is None:
        raise TypeError("H_reduced.QN_basis is None")

    if qn_compact is True:
        _qn_compact = generate_qn_compact(transitions, H_reduced)
    elif qn_compact is False:
        _qn_compact = None
    else:
        _qn_compact = qn_compact

    ground_states = H_reduced.X_states
    excited_states = H_reduced.B_states
    QN = H_reduced.QN
    H_int = H_reduced.H_int

    V_ref_int = H_reduced.V_ref_int
    if verbose:
        logger.info(
            "generate_OBE_system: 2/5 -> "
            "Generating the couplings corresponding to the transitions"
        )
    couplings = []
    for transition_selector in transition_selectors:
        if (
            transition_selector.ground_main is not None
            and transition_selector.excited_main is not None
        ):
            couplings.append(
                couplings_tlf.generate_coupling_field(
                    transition_selector.ground_main,
                    transition_selector.excited_main,
                    transition_selector.ground,
                    transition_selector.excited,
                    H_reduced.QN_basis,
                    H_int,
                    QN,
                    V_ref_int,
                    pol_vecs=transition_selector.polarizations,
                    pol_main=transition_selector.polarizations[0],
                    normalize_pol=normalize_pol,
                )
            )
        else:
            couplings.append(
                couplings_tlf.generate_coupling_field_automatic(
                    transition_selector.ground,
                    transition_selector.excited,
                    H_reduced.QN_basis,
                    H_int,
                    QN,
                    V_ref_int,
                    pol_vecs=transition_selector.polarizations,
                )
            )

    if verbose:
        logger.info("generate_OBE_system: 3/5 -> Generating the symbolic Hamiltonian")
    if _qn_compact is not None:
        H_symbolic, QN_compact = generate_total_symbolic_hamiltonian(
            QN, H_int, couplings, transition_selectors, qn_compact=_qn_compact
        )
        couplings_compact = [
            compact_coupling_field(coupling, QN, _qn_compact) for coupling in couplings
        ]
    else:
        H_symbolic = generate_total_symbolic_hamiltonian(
            QN, H_int, couplings, transition_selectors
        )

    if verbose:
        logger.info("generate_OBE_system: 4/5 -> Generating the collapse matrices")
    C_array = couplings_tlf.collapse_matrices(
        QN, ground_states, excited_states, gamma=Γ, qn_compact=_qn_compact
    )
    if decay_channels is not None:
        if isinstance(decay_channels, list):
            _decay_channels = decay_channels
        elif not isinstance(decay_channels, (tuple, list, np.ndarray)):
            _decay_channels = [decay_channels]
        elif isinstance(decay_channels, (tuple, np.ndarray)):
            _decay_channels = list(decay_channels)
        else:
            raise AssertionError(
                f"decay_channels is type f{type(decay_channels)}; supply a list, tuple"
                " or np.ndarray"
            )

        indices = utils_decay.get_insert_level_indices(
            _decay_channels, QN, excited_states
        )

        couplings = [
            insert_levels_coupling_field(coupling, indices_insert=indices)
            for coupling in couplings
        ]
        QN_decay = utils_decay.add_states_QN(_decay_channels, QN, indices)

        if _qn_compact is not None:
            indices, H_symbolic = utils_decay.add_levels_symbolic_hamiltonian(
                H_symbolic, _decay_channels, QN_compact, excited_states
            )
            C_array = utils_decay.add_decays_C_arrays(
                _decay_channels, indices, QN_compact, C_array, Γ
            )

            QN_compact = utils_decay.add_states_QN(_decay_channels, QN_compact, indices)

            couplings_compact = [
                insert_levels_coupling_field(coupling, indices_insert=indices)
                for coupling in couplings_compact
            ]
        else:
            indices, H_symbolic = utils_decay.add_levels_symbolic_hamiltonian(
                H_symbolic, _decay_channels, QN, excited_states
            )
            C_array = utils_decay.add_decays_C_arrays(
                _decay_channels, indices, QN, C_array, Γ
            )
        QN = QN_decay

    if verbose:
        logger.info(
            "generate_OBE_system: 5/5 -> Transforming the Hamiltonian and collapse "
            "matrices into a symbolic system of equations"
        )
        logging.basicConfig(level=logging.WARNING)
    ham, dissipator = generate_system_of_equations_symbolic(
        H_symbolic, C_array, fast=True, split_output=True
    )
    system = ham + dissipator

    obe_system = OBESystem(
        QN=QN_compact if _qn_compact is not None else QN,
        ground=ground_states,
        excited=excited_states,
        couplings=couplings_compact if _qn_compact is not None else couplings,
        H_symbolic=H_symbolic,
        H_int=H_int,
        V_ref_int=V_ref_int,
        C_array=C_array,
        system=system,
        dissipator=dissipator,
        coupling_symbols=[trans.Ω for trans in transition_selectors],
        polarization_symbols=[
            trans.polarization_symbols for trans in transition_selectors
        ],
        QN_original=None if _qn_compact is None else QN,
        decay_channels=_decay_channels if decay_channels else None,
        couplings_original=None if _qn_compact is None else couplings,
    )
    return obe_system


def setup_OBE_system(
    X_states: Union[states.QuantumSelector, Sequence[states.QuantumSelector]],
    B_states: Union[states.QuantumSelector, Sequence[states.QuantumSelector]],
    transitions: Sequence[couplings_tlf.TransitionSelector],
    qn_compact: Optional[
        Union[Sequence[states.QuantumSelector], states.QuantumSelector]
    ] = None,
    decay_channels: Optional[
        Union[Sequence[utils_decay.DecayChannel], utils_decay.DecayChannel]
    ] = None,
    E: npt.NDArray[np.floating] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.floating] = np.array([0.0, 0.0, 1e-5]),
    X_constants: constants.XConstants = constants.XConstants(),
    B_constants: constants.BConstants = constants.BConstants(),
    nuclear_spins: constants.TlFNuclearSpins = constants.TlFNuclearSpins(),
    Jmin_X: Optional[int] = None,
    Jmax_X: Optional[int] = None,
    Jmin_B: Optional[int] = None,
    Jmax_B: Optional[int] = None,
    transform: Optional[npt.NDArray[np.complex128]] = None,
    H_func_X: Optional[Callable] = None,
    H_func_B: Optional[Callable] = None,
    verbose: bool = False,
    normalize_pol: bool = False,
):
    """Convenience function for generating the OBE system

    Args:
        X_states : X states to include in the OBE system
        B_states : B states to include in the OBE system
        ode_parameters (odeParameters): dataclass containing the ode parameters.
                                        e.g. Ω, δ, vz, ..., etc.
        transitions (TransitionSelector): object containing all information
                                            required to generate the coupling
                                            matrices and symbolic matrix for
                                            each transition
        qn_compact (QuantumSelector): dataclass specifying a subset of states to
                                        select based on the quantum numbers
        full_output (bool, optional): Returns all matrices, states etc. if True,
                                        Returns only QN if False.
                                        Defaults to False.
        decay_channels (DecayChannel): dataclass specifying the decay channel to
                                        add
        verbose (bool, optional): Log progress to INFO. Defaults to False.

    Returns:
        full_output == True:
            list: list of states in system
        full_output == False:
            OBESystem: dataclass designed to hold the generated values
                        ground, exxcited, QN, H_int, V_ref_int, couplings,
                        H_symbolic, C_array, system
    """
    obe_system = generate_OBE_system(
        X_states,
        B_states,
        transitions,
        qn_compact=qn_compact,
        decay_channels=decay_channels,
        E=E,
        B=B,
        X_constants=X_constants,
        B_constants=B_constants,
        nuclear_spins=nuclear_spins,
        Jmin_X=Jmin_X,
        Jmax_X=Jmax_X,
        Jmin_B=Jmin_B,
        Jmax_B=Jmax_B,
        transform=transform,
        H_func_X=H_func_X,
        H_func_B=H_func_B,
        verbose=verbose,
        normalize_pol=normalize_pol,
    )
    return obe_system


def setup_OBE_system_transitions(
    transitions: Sequence[Union[OpticalTransition, MicrowaveTransition]],
    transition_selectors: Sequence[couplings_tlf.TransitionSelector],
    qn_compact: Optional[
        Union[Sequence[states.QuantumSelector], states.QuantumSelector, bool]
    ] = None,
    decay_channels: Optional[
        Union[Sequence[utils_decay.DecayChannel], utils_decay.DecayChannel]
    ] = None,
    E: npt.NDArray[np.floating] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.floating] = np.array([0.0, 0.0, 1e-5]),
    X_constants: constants.XConstants = constants.XConstants(),
    B_constants: constants.BConstants = constants.BConstants(),
    nuclear_spins: constants.TlFNuclearSpins = constants.TlFNuclearSpins(),
    Jmin_X: Optional[int] = None,
    Jmax_X: Optional[int] = None,
    Jmin_B: Optional[int] = None,
    Jmax_B: Optional[int] = None,
    transform: Optional[npt.NDArray[np.complex128]] = None,
    H_func_X: Optional[Callable] = None,
    H_func_B: Optional[Callable] = None,
    verbose: bool = False,
    normalize_pol: bool = False,
    Γ: float = hamiltonian.Γ,
) -> OBESystem:
    """Convenience function for generating the OBE system

    Args:
        ode_parameters (odeParameters): dataclass containing the ode parameters.
                                        e.g. Ω, δ, vz, ..., etc.
        transitions (Sequence[TransitionSelector]): Sequence containing all transition
                                            information required to generate
                                            the coupling matrices and symbolic matrix
                                            for each transition
        qn_compact (QuantumSelector): dataclass specifying a subset of states to
                                        select based on the quantum numbers
        full_output (bool, optional): Returns all matrices, states etc. if True,
                                        Returns only QN if False.
                                        Defaults to False.
        decay_channels (DecayChannel): dataclass specifying the decay channel to
                                        add
        verbose (bool, optional): Log progress to INFO. Defaults to False.

    Returns:
        full_output == True:
            list: list of states in system
        full_output == False:
            OBESystem: dataclass designed to hold the generated values
                        ground, exxcited, QN, H_int, V_ref_int, couplings,
                        H_symbolic, C_array, system
    """
    obe_system = generate_OBE_system_transitions(
        transitions=transitions,
        transition_selectors=transition_selectors,
        qn_compact=qn_compact,
        decay_channels=decay_channels,
        E=E,
        B=B,
        Γ=Γ,
        X_constants=X_constants,
        B_constants=B_constants,
        nuclear_spins=nuclear_spins,
        Jmin_X=Jmin_X,
        Jmax_X=Jmax_X,
        Jmin_B=Jmin_B,
        Jmax_B=Jmax_B,
        transform=transform,
        H_func_X=H_func_X,
        H_func_B=H_func_B,
        verbose=verbose,
        normalize_pol=normalize_pol,
    )

    return obe_system



"""Lambda-doubling Hamiltonian operators for B state in coupled Omega basis.

This module implements lambda-doubling interactions in the B electronic state,
including the q-term (Λ-doubling) and nuclear spin-rotation coupling that couples
states with opposite Omega quantum numbers.

The lambda-doubling effect arises from the interaction between electronic and
rotational motion in molecules with non-zero electronic angular momentum.
"""

from __future__ import annotations

from functools import lru_cache

import numpy as np

from centrex_tlf.constants import BConstants
from centrex_tlf.states import CoupledBasisState, CoupledState

from ..wigner import sixj_f, threej_f


@lru_cache(maxsize=int(1e6))
def H_LD(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """Lambda-doubling q-term operator for B state in coupled Omega basis.

    Calculates the lambda-doubling "q-term" that couples states with opposite Omega
    values, shifting e-parity levels up and f-parity levels down in energy. This term
    arises from the interaction between electronic orbital motion and molecular rotation.

    The matrix element is diagonal in all quantum numbers except Omega, which changes
    sign: |J,F₁,F,mF,Ω⟩ → |J,F₁,F,mF,-Ω⟩

    Args:
        psi: Input coupled basis state |J,F₁,F,mF,I₁,I₂,Ω⟩ in B electronic state.
        constants: Molecular constants for TlF B state, requires q parameter.

    Returns:
        CoupledState containing the lambda-doubling contribution with amplitude
        q·J(J+1)/2 and basis state with inverted Omega.

    Notes:
        - The operator is off-diagonal in Omega (changes sign)
        - The energy shift is proportional to J(J+1)
        - e-parity levels shift up, f-parity levels shift down

    Example:
        >>> state = CoupledBasisState(F=1, mF=0, F1=0.5, J=1, I1=0.5, I2=0.5, Omega=1)
        >>> constants = BConstants(q=2423000.0)
        >>> result = H_LD(state, constants)
        >>> # Returns state with Omega=-1 and amplitude q·J(J+1)/2
    """
    # All quantum numbers the same, except Omega inverts sign
    J = psi.J
    I1 = psi.I1
    I2 = psi.I2
    F1 = psi.F1
    F = psi.F
    mF = psi.mF
    Omega = -psi.Omega

    amp = constants.q * J * (J + 1) / 2
    ket = CoupledBasisState(
        F,
        mF,
        F1,
        J,
        I1,
        I2,
        Omega=Omega,
        electronic_state=psi.electronic_state,
        P=psi.P,
    )

    return CoupledState([(amp, ket)])


@lru_cache(maxsize=int(1e6))
def H_cp1_Tl(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """Lambda-doubling nuclear spin-rotation operator for Tl nucleus in B state.

    Calculates the lambda-doubling nuclear spin-rotation interaction term (c'₁) for
    the thallium (Tl) nucleus. This term couples states with different J values and
    opposite Omega values, arising from the interaction between the Tl nuclear spin
    and the molecular rotation in the presence of lambda-doubling.

    This operator is off-diagonal in both J (ΔJ = 0, ±1) and Omega (changes sign).
    The quantum numbers I₁, I₂, F₁, F, and mF remain unchanged.

    Args:
        psi: Input coupled basis state |J,F₁,F,mF,I₁,I₂,Ω⟩ in B electronic state.
        constants: Molecular constants for TlF B state, requires c1p_Tl parameter
            (Tl lambda-doubling nuclear spin-rotation constant).

    Returns:
        CoupledState containing sum over all allowed transitions with ΔJ = 0, ±1
        and Omega → -Omega. Each term includes appropriate Wigner 6-j and 3-j
        symbols and angular momentum factors.

    Notes:
        - Couples states with J' = J-1, J, J+1 and Omega' = -Omega
        - The matrix element involves two terms (T(J)T(J) and T(Jp)T(Jp))
        - Typical values for c1p_Tl in TlF B state: ~11 MHz
        - Selection rules: ΔJ = 0, ±1; ΔΩ = 0 but Omega changes sign

    References:
        Brown, J. M., & Carrington, A. (2003). "Rotational Spectroscopy of Diatomic
        Molecules." Cambridge University Press. Chapter on lambda-doubling.

    Example:
        >>> state = CoupledBasisState(F=1, mF=0, F1=0.5, J=1, I1=0.5, I2=0.5, Omega=1)
        >>> constants = BConstants(c1p_Tl=11170000.0)
        >>> result = H_cp1_Tl(state, constants)
        >>> # Returns superposition of states with J=0,1,2 and Omega=-1
    """
    # Find the quantum numbers of the input state
    Jp = psi.J
    I1p = psi.I1
    I2p = psi.I2
    F1p = psi.F1
    Fp = psi.F
    mFp = psi.mF
    Omegap = psi.Omega

    # I1, I2, F and mF are the same for both states
    I1 = I1p
    I2 = I2p
    F = Fp
    F1 = F1p
    mF = mFp

    # Omegas are opposite
    Omega = -Omegap

    # Calculate the value of q
    q = Omega

    data = []

    # Loop over possible values of J
    # Need J = |Jp-1| ... |Jp+1|
    for J in np.arange(np.abs(Jp - 1), Jp + 2):
        # Calculate matrix element
        amp = (
            -constants.c1p_Tl
            / 2
            * (-1) ** (J + Jp + F1 + I1 - Omegap)
            * sixj_f(I1, Jp, F1, J, I1, 1)
            * np.sqrt(I1 * (I1 + 1) * (2 * I1 + 1) * (2 * J + 1) * (2 * Jp + 1))
            * (
                (-1) ** (J)
                * np.sqrt(J * (J + 1) * (2 * J + 1))
                * threej_f(J, 1, Jp, 0, q, Omegap)
                * threej_f(J, 1, J, -Omega, q, 0)
                + (-1) ** (Jp)
                * np.sqrt(Jp * (Jp + 1) * (2 * Jp + 1))
                * threej_f(Jp, 1, Jp, 0, q, Omegap)
                * threej_f(J, 1, Jp, -Omega, q, 0)
            )
        )

        basis_state = CoupledBasisState(
            F,
            mF,
            F1,
            J,
            I1,
            I2,
            Omega=Omega,
            electronic_state=psi.electronic_state,
            P=psi.P,
        )

        data.append((amp, basis_state))

    return CoupledState(data)



"""Doppler shift and velocity-to-detuning conversion utilities.

This module provides functions to calculate Doppler-shifted frequencies and
convert molecular velocities to laser detunings for spectroscopy applications.
"""
from typing import Union, overload

import numpy as np
import numpy.typing as npt
import scipy.constants as cst

__all__ = ["doppler_shift", "velocity_to_detuning"]

# Define type aliases for clarity
FloatOrArray = Union[float, npt.NDArray[np.floating]]


@overload
def doppler_shift(velocity: float, frequency: float = 1.1e15) -> float: ...
@overload
def doppler_shift(
    velocity: npt.NDArray[np.floating], frequency: float = 1.1e15
) -> npt.NDArray[np.floating]: ...
@overload
def doppler_shift(
    velocity: float, frequency: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]: ...
@overload
def doppler_shift(
    velocity: npt.NDArray[np.floating], frequency: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]: ...


def doppler_shift(
    velocity: FloatOrArray, frequency: FloatOrArray = 1.1e15
) -> FloatOrArray:
    """Calculate the Doppler-shifted frequency for a given velocity.

    Uses the non-relativistic Doppler formula: f' = f(1 + v/c) where c is the speed
    of light. This approximation is valid for v << c.

    Args:
        velocity (FloatOrArray): Velocity in m/s (float or array). Positive for
            molecules approaching the laser (blue-shift), negative for receding
            (red-shift).
        frequency (FloatOrArray): Rest-frame frequency in Hz (float or array).
            Defaults to 1.1e15 Hz (TlF B-X transition at ~272 nm).

    Returns:
        FloatOrArray: Doppler-shifted frequency in Hz. Returns same type as input
            (float or array).

    Raises:
        ValueError: If frequency is non-positive.

    Example:
        >>> doppler_shift(100.0)  # 100 m/s towards observer
        1.1000003666666666e+15
        >>> doppler_shift(-100.0)  # 100 m/s away from observer
        1.0999996333333334e+15
        >>> doppler_shift(np.array([0, 100, -100]))  # Array of velocities
        array([1.1e+15, 1.10000037e+15, 1.09999963e+15])
        
    Note:
        For relativistic velocities (v > 0.1c), use the relativistic Doppler formula.
    """
    if np.any(np.asarray(frequency) <= 0):
        raise ValueError(f"Frequency must be positive, got {frequency}")

    return frequency * (1 + velocity / cst.c)


@overload
def velocity_to_detuning(velocity: float, frequency: float = 1.1e15) -> float: ...
@overload
def velocity_to_detuning(
    velocity: npt.NDArray[np.floating], frequency: float = 1.1e15
) -> npt.NDArray[np.floating]: ...
@overload
def velocity_to_detuning(
    velocity: float, frequency: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]: ...
@overload
def velocity_to_detuning(
    velocity: npt.NDArray[np.floating], frequency: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]: ...


def velocity_to_detuning(
    velocity: FloatOrArray, frequency: FloatOrArray = 1.1e15
) -> FloatOrArray:
    """Convert molecular velocity to laser detuning based on Doppler shift.

    Calculates the detuning from resonance: Δω = ω(v/c) where ω = 2πf is the angular
    frequency. This represents the frequency shift in the molecular frame.

    Args:
        velocity (FloatOrArray): Velocity in m/s (float or array). Positive for
            molecules approaching the laser (positive detuning), negative for receding
            (negative detuning).
        frequency (FloatOrArray): Rest-frame transition frequency in Hz (float or array).
            Defaults to 1.1e15 Hz (TlF B-X transition at ~272 nm).

    Returns:
        FloatOrArray: Detuning in rad/s. Positive means laser is blue-detuned in
            molecular frame, negative means red-detuned. Returns same type as input.

    Raises:
        ValueError: If frequency is non-positive.

    Example:
        >>> velocity_to_detuning(100.0)  # 100 m/s molecular velocity
        2.3089690054388264e+09
        >>> velocity_to_detuning(np.array([50, 100, 150]))  # Multiple velocities
        array([1.15448450e+09, 2.30896901e+09, 3.46345351e+09])
        
    Note:
        The detuning Δω = ω' - ω₀ where ω' is the laser frequency in the molecular
        frame and ω₀ is the atomic transition frequency.
    """
    if np.any(np.asarray(frequency) <= 0):
        raise ValueError(f"Frequency must be positive, got {frequency}")

    # Direct computation of detuning in rad/s
    return frequency * (velocity / cst.c) * (2 * np.pi)



import numpy as np

from centrex_tlf.constants import BConstants
from centrex_tlf.states import UncoupledBasisState, UncoupledState

from .wigner import threej_f

__all__ = ["H_LD", "H_c1p", "H_mhf_Tl", "H_mhf_F", "HZx", "HZy", "HZz"]


########################################################
# Λ doubling term
########################################################


def H_LD(psi: UncoupledBasisState, coefficients: BConstants) -> UncoupledState:
    """Lambda-doubling Hamiltonian for B state in uncoupled basis.
    
    Mixes states with opposite Ω (Ω → -Ω). Same physics as coupled basis version.
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
        coefficients (BConstants): B state molecular constants (q parameter)
    
    Returns:
        UncoupledState: Lambda-doubling contribution in uncoupled basis
    """
    J = psi.J
    mJ = psi.mJ
    I1 = psi.I1
    m1 = psi.m1
    I2 = psi.I2
    m2 = psi.m2
    Omega = psi.Omega
    if Omega is None:
        raise TypeError("can only have Omega be int (not None)")
    Omegaprime = -Omega

    amp = (
        coefficients.q
        * (-1) ** (J - Omegaprime)
        / (2 * np.sqrt(6))
        * threej_f(J, 2, J, -Omegaprime, Omegaprime - Omega, Omega)
        * np.sqrt((2 * J - 1) * 2 * J * (2 * J + 1) * (2 * J + 2) * (2 * J + 3))
    )
    ket = UncoupledBasisState(
        J, mJ, I1, m1, I2, m2, Omegaprime, electronic_state=psi.electronic_state
    )

    return UncoupledState([(amp, ket)])


########################################################
# C'(Tl) - term (Brown 1978 "A determination of fundamental Zeeman parameters
# for the OH radical", eqn A12)
########################################################


def H_c1p(psi: UncoupledBasisState, coefficients: BConstants) -> UncoupledState:
    """Nuclear spin-rotation interaction C'(Tl) for B state in uncoupled basis.
    
    Ref: Brown 1978 "A determination of fundamental Zeeman parameters for the OH
    radical", equation A12.
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
        coefficients (BConstants): B state molecular constants (cp1_Tl)
    
    Returns:
        UncoupledState: Nuclear spin-rotation contribution for Tl in uncoupled basis
    """
    # Find the quantum numbers of the input state
    J = psi.J
    mJ = psi.mJ
    I1 = psi.I1
    m1 = psi.m1
    I2 = psi.I2
    m2 = psi.m2
    Omega = psi.Omega
    if Omega is None:
        raise TypeError("can only have Omega be int (not None)")

    # I1, I2 and m2 must be the same for non-zero matrix element
    I1prime = I1
    m2prime = m2
    I2prime = I2

    # To have non-zero matrix element need OmegaPrime = -Omega
    Omegaprime = -Omega

    # q is chosen such that q == Omegaprime
    q = Omega

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over the possible values of quantum numbers for which the matrix element can
    # be non-zero
    # Need Jprime = J+1 ... |J-1|
    for Jprime in range(np.abs(J - 1), J + 2):
        # Loop over possible values of mJprime and m1prime
        for mJprime in range(-Jprime, Jprime + 1):
            # Must have mJ+m1 = mJprime + m1prime
            m1prime = mJ + m1 - mJprime
            if np.abs(m1prime <= I1):
                # Evaluate the matrix element

                # Matrix element for T(J)T(I)
                term1 = (
                    (-1) ** (Jprime - Omegaprime + I1 - m1 - q + mJprime)
                    * np.sqrt(
                        Jprime
                        * (Jprime + 1)
                        * (2 * Jprime + 1) ** 2
                        * (2 * J + 1)
                        * I1
                        * (I1 + 1)
                        * (2 * I1 + 1)
                    )
                    * (threej_f(Jprime, 1, J, -mJprime, mJprime - mJ, mJ))
                    * (threej_f(I1, 1, I1, -m1prime, m1prime - m1, m1))
                    * (threej_f(Jprime, 1, J, 0, -q, Omega))
                    * (threej_f(Jprime, 1, Jprime, -Omegaprime, -q, 0))
                )

                # Matrix element for T(I)T(J)
                term2 = (
                    (-1) ** (mJprime + J - Omegaprime + I1 - m1 - q)
                    * np.sqrt(
                        J
                        * (J + 1)
                        * (2 * J + 1) ** 2
                        * (2 * Jprime + 1)
                        * I1
                        * (I1 + 1)
                        * (2 * I1 + 1)
                    )
                    * (threej_f(Jprime, 1, J, -mJprime, mJprime - mJ, mJ))
                    * (threej_f(Jprime, 1, J, -Omegaprime, -q, 0))
                    * (threej_f(J, 1, J, 0, -q, Omega))
                    * (threej_f(I1, 1, I1, -m1prime, m1prime - m1, m1))
                )

                amp = coefficients.c_Tl * 0.5 * (term1 + term2)

                basis_state = UncoupledBasisState(
                    Jprime,
                    mJprime,
                    I1prime,
                    m1prime,
                    I2prime,
                    m2prime,
                    Omegaprime,
                    P=psi.P,
                    electronic_state=psi.electronic_state,
                )

                if amp != 0:
                    data.append((amp, basis_state))

    return UncoupledState(data)


########################################################
# Electron magnetic hyperfine operator
########################################################


def H_mhf_Tl(psi: UncoupledBasisState, coefficients: BConstants) -> UncoupledState:
    """Magnetic hyperfine interaction for Tl nucleus in B state (uncoupled basis).
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
        coefficients (BConstants): B state molecular constants (h1_Tl)
    
    Returns:
        UncoupledState: Magnetic hyperfine contribution for Tl in uncoupled basis
    """
    # Find the quantum numbers of the input state
    J = psi.J
    mJ = psi.mJ
    I1 = psi.I1
    m1 = psi.m1
    I2 = psi.I2
    m2 = psi.m2
    Omega = psi.Omega
    if Omega is None:
        raise TypeError("can only have Omega be int (not None)")

    # I1, I2 and m2 must be the same for non-zero matrix element
    I2prime = I2
    m2prime = m2
    I1prime = I1

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over the possible values of quantum numbers for which the matrix element can
    # be non-zero
    # Need Jprime = J+1 ... |J-1|
    for Jprime in range(np.abs(J - 1), J + 2):
        # Evaluate the part of the matrix element that is common for all p
        common_coefficient = (
            coefficients.h1_Tl
            * threej_f(J, 1, Jprime, -Omega, 0, Omega)
            * np.sqrt((2 * J + 1) * (2 * Jprime + 1) * I1 * (I1 + 1) * (2 * I1 + 1))
        )

        # Loop over the spherical tensor components of I1:
        for p in range(-1, 2):
            # To have non-zero matrix element need mJ-p = mJprime
            mJprime = mJ + p

            # Also need m2 - p = m2prime
            m1prime = m1 - p

            # Check that mJprime and m2prime are physical
            if np.abs(mJprime) <= Jprime and np.abs(m1prime) <= I1prime:
                # Calculate rest of matrix element
                p_factor = (
                    (-1) ** (p - mJ + I1 - m1 - Omega)
                    * threej_f(J, 1, Jprime, -mJ, -p, mJprime)
                    * threej_f(I1, 1, I1prime, -m1, p, m1prime)
                )

                amp = Omega * common_coefficient * p_factor
                basis_state = UncoupledBasisState(
                    Jprime, mJprime, I1prime, m1prime, I2prime, m2prime, psi.Omega
                )
                if amp != 0:
                    data.append((amp, basis_state))

    return UncoupledState(data)


def H_mhf_F(psi: UncoupledBasisState, coefficients: BConstants) -> UncoupledState:
    """Magnetic hyperfine interaction for F nucleus in B state (uncoupled basis).
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
        coefficients (BConstants): B state molecular constants (h1_F)
    
    Returns:
        UncoupledState: Magnetic hyperfine contribution for F in uncoupled basis
    """
    # Find the quantum numbers of the input state
    J = psi.J
    mJ = psi.mJ
    I1 = psi.I1
    m1 = psi.m1
    I2 = psi.I2
    m2 = psi.m2
    Omega = psi.Omega
    if Omega is None:
        raise TypeError("can only have Omega be int (not None)")

    # I1, I2 and m1 must be the same for non-zero matrix element
    I1prime = I1
    m1prime = m1
    I2prime = I2

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over the possible values of quantum numbers for which the matrix element can
    # be non-zero
    # Need Jprime = J+1 ... |J-1|
    for Jprime in range(np.abs(J - 1), J + 2):
        # Evaluate the part of the matrix element that is common for all p
        common_coefficient = (
            coefficients.h1_F
            * threej_f(J, 1, Jprime, -Omega, 0, Omega)
            * np.sqrt((2 * J + 1) * (2 * Jprime + 1) * I2 * (I2 + 1) * (2 * I2 + 1))
        )

        # Loop over the spherical tensor components of I2:
        for p in range(-1, 2):
            # To have non-zero matrix element need mJ-p = mJprime
            mJprime = mJ + p

            # Also need m2 - p = m2prime
            m2prime = m2 - p

            # Check that mJprime and m2prime are physical
            if np.abs(mJprime) <= Jprime and np.abs(m2prime) <= I2prime:
                # Calculate rest of matrix element
                p_factor = (
                    (-1) ** (p - mJ + I2 - m2 - Omega)
                    * threej_f(J, 1, Jprime, -mJ, -p, mJprime)
                    * threej_f(I2, 1, I2prime, -m2, p, m2prime)
                )

                amp = Omega * common_coefficient * p_factor
                basis_state = UncoupledBasisState(
                    Jprime, mJprime, I1prime, m1prime, I2prime, m2prime, psi.Omega
                )
                if amp != 0:
                    data.append((amp, basis_state))

    return UncoupledState(data)


########################################################
# Zeeman B state Hamiltonian
########################################################


def HZx(psi: UncoupledBasisState) -> UncoupledState:
    """Zeeman Hamiltonian for Bₓ field component in B state (uncoupled basis).
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
    
    Returns:
        UncoupledState: Zeeman interaction for Bₓ (currently placeholder)
    
    Note:
        TODO: Full implementation needed
    """
    # TODO
    return UncoupledState([(1.0, psi)])


def HZy(psi: UncoupledBasisState) -> UncoupledState:
    """Zeeman Hamiltonian for Bᵧ field component in B state (uncoupled basis).
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
    
    Returns:
        UncoupledState: Zeeman interaction for Bᵧ (currently placeholder)
    
    Note:
        TODO: Full implementation needed
    """
    # TODO
    return UncoupledState([(1.0, psi)])


def HZz(psi: UncoupledBasisState, coefficients: BConstants) -> UncoupledState:
    """Zeeman Hamiltonian for Bz field component in B state (uncoupled basis).
    
    Args:
        psi (UncoupledBasisState): Uncoupled basis state |J,mJ,I₁,m₁,I₂,m₂,Ω⟩
        coefficients (BConstants): B state molecular constants (μ_B Bohr magneton)
    
    Returns:
        UncoupledState: Zeeman interaction for Bz field component
    """
    # Find the quantum numbers of the input state
    J = psi.J
    mJ = psi.mJ
    I1 = psi.I1
    m1 = psi.m1
    I2 = psi.I2
    m2 = psi.m2
    Omega = psi.Omega
    if Omega is None:
        raise TypeError("can only have Omega be int (not None)")
    S = 1

    # The other state must have the same value for I1,m1,I2,m2,mJ and Omega
    I1prime = I1
    m1prime = m1
    I2prime = I2
    m2prime = m2
    Omegaprime = Omega
    mJprime = mJ

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over possible values of Jprime
    for Jprime in range(np.abs(J - 1), J + 2):
        # Electron orbital angular momentum term
        L_term = (
            coefficients.gL
            * Omega
            * np.sqrt((2 * J + 1) * (2 * Jprime + 1))
            * (-1) ** (mJprime - Omegaprime)
            * (threej_f(Jprime, 1, J, -mJprime, 0, mJ))
            * (threej_f(Jprime, 1, J, -Omegaprime, 0, Omega))
        )

        # Electron spin term
        S_term = (
            coefficients.gS
            * np.sqrt((2 * J + 1) * (2 * Jprime + 1))
            * (-1) ** (mJprime - Omegaprime)
            * (threej_f(Jprime, 1, J, -mJprime, 0, mJ))
            * (threej_f(Jprime, 1, J, -Omegaprime, 0, Omega))
            * (-1) ** (S)
            * (threej_f(S, 1, S, 0, 0, 0))
            * np.sqrt(S * (S + 1) * (2 * S + 1))
        )

        amp = L_term + S_term
        basis_state = UncoupledBasisState(
            Jprime, mJprime, I1prime, m1prime, I2prime, m2prime, Omegaprime
        )

        if amp != 0:
            data.append((amp, basis_state))

    return UncoupledState(data)



from . import (
    branching,
    collapse,
    coupling_matrix,
    polarization,
    transition,
    utils,
    utils_compact,
)
from .branching import *  # noqa
from .collapse import *  # noqa
from .coupling_matrix import *  # noqa
from .polarization import *  # noqa
from .transition import *  # noqa
from .utils import *  # noqa
from .utils_compact import *  # noqa

__all__ = branching.__all__.copy()
__all__ += collapse.__all__.copy()
__all__ += coupling_matrix.__all__.copy()
__all__ += polarization.__all__.copy()
__all__ += transition.__all__.copy()
__all__ += utils.__all__.copy()
__all__ += utils_compact.__all__.copy()



import matplotlib.pyplot as plt
import numpy as np

from centrex_tlf import hamiltonian, states

# generate the hyperfine sublevels in J=0 to J=4
QN = states.generate_coupled_states_excited(
    Js=np.arange(1, 5), Ps=None, Omegas=[-1, 1], basis=states.Basis.CoupledΩ
)


# generate the X hamiltonian terms
H = hamiltonian.generate_coupled_hamiltonian_B(QN)

# create a function outputting the hamiltonian as a function of E and B
Hfunc = hamiltonian.generate_coupled_hamiltonian_B_function(H)

# V/cm
Ez = np.linspace(-1_000, 1_000, 101)

# generate the Hamiltonian for (almost) zero field, add a small field to make states
# non-degenerate
Hi = Hfunc(E=[0, 0, 1e-3], B=[0, 0, 1e-3])
E, V = np.linalg.eigh(Hi)

# get the true superposition-states of the system
QN_states = hamiltonian.matrix_to_states(V, QN)

# original eigenvectors used in tracking states as energies change order
V_track = V.copy()

# empty array for storing energies
energy = np.empty([Ez.size, len(QN)], dtype=np.float64)

# iterate over the electric field values
for idx, Ei in enumerate(Ez):
    Hi = Hfunc(E=[0.0, 0.0, Ei], B=[0.0, 0.0, 0.0])
    E, V = np.linalg.eigh(Hi)

    # sort indices to keep the state order the same
    indices = np.argmax(np.abs(V_track.conj().T @ V), axis=1)
    energy[idx, :] = E[indices]
    V_track[:, :] = V[:, indices]

# indices of the J'=1 states
indices_J1 = [idx for idx, s in enumerate(QN_states) if s.largest.J == 1]


# plot the J'=1 Zeeman Curves
fig, ax = plt.subplots(figsize=(12, 8))
ax.plot(
    Ez,
    (energy.real[:, indices_J1] - energy.real[:, indices_J1][0, 0]) / (2 * np.pi * 1e6),
    lw=3,
    color="k",
)
ax.set_xlabel("E [V/cm]")
ax.set_ylabel("Energy [MHz]")
ax.set_title("|J'=1> Stark Curves")
ax.grid(True)
plt.show()



from centrex_tlf import couplings, states


def test_check_transition_coupled_allowed():
    ground_state = states.CoupledBasisState(
        J=1,
        F1=1 / 2,
        F=1,
        mF=0,
        I1=1 / 2,
        I2=1 / 2,
        Omega=0,
        P=-1,
        electronic_state=states.ElectronicState.X,
    )
    excited_state = states.CoupledBasisState(
        J=1,
        F1=1 / 2,
        F=1,
        mF=0,
        I1=1 / 2,
        I2=1 / 2,
        Omega=1,
        P=1,
        electronic_state=states.ElectronicState.B,
    )
    allowed = couplings.utils.check_transition_coupled_allowed_polarization(
        ground_state, excited_state, ΔmF_allowed=0, return_err=False
    )
    if isinstance(allowed, bool):
        assert not allowed

    excited_state.mF = -1
    allowed = couplings.utils.check_transition_coupled_allowed_polarization(
        ground_state, excited_state, ΔmF_allowed=0, return_err=False
    )
    if isinstance(allowed, bool):
        assert not allowed

    allowed = couplings.utils.check_transition_coupled_allowed_polarization(
        ground_state, excited_state, ΔmF_allowed=-1, return_err=False
    )
    print(allowed)
    if isinstance(allowed, bool):
        assert allowed

    excited_state.mF = 1
    allowed = couplings.utils.check_transition_coupled_allowed_polarization(
        ground_state, excited_state, ΔmF_allowed=1, return_err=False
    )
    if isinstance(allowed, bool):
        assert allowed

    excited_state.mF = 1
    allowed = couplings.utils.check_transition_coupled_allowed_polarization(
        ground_state, excited_state, ΔmF_allowed=-1, return_err=False, ΔmF_absolute=True
    )
    if isinstance(allowed, bool):
        assert allowed



"""Rabi frequency calculations and beam parameter conversions.

This module provides utilities for converting between laser power, intensity,
electric field, and Rabi frequencies for both optical and microwave transitions.
Includes Gaussian and rectangular beam profiles, and beam width conversions.
"""

from typing import TypeVar

import numpy as np
import numpy.typing as npt
import scipy.constants as cst

from centrex_tlf.constants import ED_XtB, XConstants

__all__ = [
    "fwhm_to_sigma",
    "sigma_to_fwhm",
    "electric_field_to_rabi",
    "sigma_to_waist",
    "waist_to_sigma",
    "electric_field_to_intensity",
    "intensity_to_electric_field",
    "intensity_to_power_gaussian_beam",
    "intensity_to_power_rectangular_beam",
    "intensity_to_rabi",
    "power_to_intensity_gaussian_beam",
    "power_to_intensity_rectangular_beam",
    "power_to_rabi_gaussian_beam",
    "power_to_rabi_gaussian_beam_microwave",
    "power_to_rabi_rectangular_beam",
    "rabi_to_power_gaussian_beam",
    "rabi_to_power_gaussian_beam_microwave",
    "rabi_to_electric_field",
]

T = TypeVar("T", float, npt.NDArray[np.floating])


def fwhm_to_sigma(fwhm: T) -> T:
    """Convert full width at half maximum (FWHM) to standard deviation.

    For a Gaussian profile: σ = FWHM / (2√(2ln2)) ≈ FWHM / 2.355.

    Args:
        fwhm (T): Full width at half maximum [arbitrary units].

    Returns:
        T: Standard deviation σ in same units as input.

    Example:
        >>> fwhm_to_sigma(2.355)  # FWHM of 2.355 gives σ ≈ 1
        1.0
    """
    return fwhm / (2 * np.sqrt(2 * np.log(2)))


def sigma_to_fwhm(sigma: T) -> T:
    """Convert standard deviation to full width at half maximum (FWHM).

    For a Gaussian profile: FWHM = σ · 2√(2ln2) ≈ σ · 2.355.

    Args:
        sigma (T): Standard deviation σ [arbitrary units].

    Returns:
        T: Full width at half maximum in same units as input.

    Example:
        >>> sigma_to_fwhm(1.0)  # σ of 1 gives FWHM ≈ 2.355
        2.3548200450309493
    """
    return sigma * 2 * np.sqrt(2 * np.log(2))


def sigma_to_waist(sigma: T) -> T:
    """Convert Gaussian standard deviation to beam waist (1/e² radius).

    For Gaussian beams: w₀ = 2σ where w₀ is the waist radius at which intensity
    drops to 1/e² of peak value.

    Args:
        sigma (T): Standard deviation σ [m].

    Returns:
        T: Beam waist w₀ in same units as input.

    Example:
        >>> sigma_to_waist(0.5e-3)  # 0.5 mm sigma gives 1 mm waist
        0.001
    """
    return sigma * 2


def waist_to_sigma(waist: T) -> T:
    """Convert beam waist (1/e² radius) to Gaussian standard deviation.

    For Gaussian beams: σ = w₀/2 where w₀ is the waist radius.

    Args:
        waist (T): Beam waist w₀ [m].

    Returns:
        T: Standard deviation σ in same units as input.

    Example:
        >>> waist_to_sigma(1e-3)  # 1 mm waist gives 0.5 mm sigma
        0.0005
    """
    return waist / 2


def intensity_to_electric_field(intensity: T) -> T:
    """Convert optical intensity to electric field amplitude.

    Uses the time-averaged electromagnetic energy density relation for plane waves:
    I = (1/2)·c·ε₀·E₀², solving for E₀ gives E₀ = √(2I/(c·ε₀)).

    Args:
        intensity (T): Optical intensity [W/m²].

    Returns:
        T: Electric field amplitude E₀ [V/m]. For time-harmonic fields,
            E(t) = E₀·cos(ωt).

    Raises:
        ValueError: If intensity is negative.

    Example:
        >>> intensity_to_electric_field(1000.0)  # 1 kW/m² intensity
        868.3641436915924

    Note:
        For focused Gaussian beams, intensity is the peak on-axis value.
    """
    if np.any(intensity < 0):
        raise ValueError("Intensity must be non-negative")
    return np.sqrt((2 / (cst.c * cst.epsilon_0)) * intensity)  # type: ignore[return-value]


def electric_field_to_rabi(electric_field: T, coupling: float, D: float) -> T:
    """Calculate Rabi frequency from electric field and transition coupling.

    Rabi frequency Ω = (E₀·μ·coupling)/ℏ where μ is the transition dipole moment
    and coupling is the reduced matrix element (e.g., Clebsch-Gordan coefficient).

    Args:
        electric_field (T): Electric field amplitude E₀ [V/m].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        D (float): Transition dipole moment [C·m].

    Returns:
        T: Rabi frequency [rad/s]. The on-resonance oscillation frequency.

    Example:
        >>> E = 1000.0  # V/m
        >>> electric_field_to_rabi(E, 1.0, 2.67e-30)  # TlF typical values
        2.532...e+05

    Note:
        For two-level systems, population oscillates at frequency Ω/2.
    """
    return electric_field * coupling * D / cst.hbar


def intensity_to_rabi(intensity: T, coupling: float, D: float) -> T:
    """Calculate Rabi frequency directly from laser intensity.

    Combines intensity-to-field and field-to-Rabi conversions:
    Ω = √(2I/(c·ε₀)) · (μ·coupling)/ℏ.

    Args:
        intensity (T): Optical intensity [W/m²].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        D (float): Transition dipole moment [C·m].

    Returns:
        T: Rabi frequency [rad/s].

    Example:
        >>> intensity_to_rabi(1000.0, 1.0, 2.67e-30)  # 1 kW/m²
        2.199...e+05

    Note:
        This is the most commonly used conversion for optical transitions.
    """
    electric_field = intensity_to_electric_field(intensity)
    rabi = electric_field_to_rabi(electric_field, coupling, D)
    return rabi


def intensity_to_power_rectangular_beam(intensity: T, wx: float, wy: float) -> T:
    """Convert intensity to total power for rectangular (flat-top) beam.

    Assumes uniform intensity across rectangular cross-section: P = I·A where A = wx·wy.

    Args:
        intensity (T): Intensity [W/m²].
        wx (float): Beam width in x-direction [m].
        wy (float): Beam width in y-direction [m].

    Returns:
        T: Total optical power [W].

    Example:
        >>> intensity_to_power_rectangular_beam(1000.0, 1e-3, 2e-3)  # 1x2 mm beam
        0.002
    """
    return intensity * wx * wy


def power_to_intensity_rectangular_beam(power: T, wx: float, wy: float) -> T:
    """Convert total power to intensity for rectangular (flat-top) beam.

    Assumes uniform intensity across rectangular cross-section: I = P/A where A = wx·wy.

    Args:
        power (T): Total optical power [W].
        wx (float): Beam width in x-direction [m].
        wy (float): Beam width in y-direction [m].

    Returns:
        T: Intensity [W/m²].

    Example:
        >>> power_to_intensity_rectangular_beam(0.001, 1e-3, 1e-3)  # 1 mW in 1x1 mm
        1000.0
    """
    return power / (wx * wy)


def power_to_rabi_rectangular_beam(
    power: T,
    coupling: float,
    wx: float,
    wy: float,
    D: float = ED_XtB,
) -> T:
    """Calculate Rabi frequency from power for rectangular beam profile.

    Combines power-to-intensity conversion (assuming flat-top beam) with Rabi
    frequency calculation. Useful for microwave transitions or multimode lasers.

    Args:
        power (T): Total optical power [W].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        wx (float): Beam width in x-direction [m].
        wy (float): Beam width in y-direction [m].
        D (float): Transition dipole moment [C·m]. Defaults to 2.6675506e-30 for
            TlF B-X optical transition.

    Returns:
        T: Rabi frequency [rad/s].

    Example:
        >>> power_to_rabi_rectangular_beam(0.001, 1.0, 1e-3, 1e-3)  # 1 mW, 1x1 mm
        2.199...e+05
    """
    intensity = power_to_intensity_rectangular_beam(power, wx, wy)
    rabi = intensity_to_rabi(intensity, coupling, D)
    return rabi


def power_to_intensity_gaussian_beam(power: T, sigma_x: float, sigma_y: float) -> T:
    """Convert total power to peak intensity for Gaussian beam.

    For Gaussian beam I(x,y) = I₀·exp(-x²/(2σₓ²)-y²/(2σᵧ²)), integrating over all
    space gives P = I₀·2π·σₓ·σᵧ, so I₀ = P/(2π·σₓ·σᵧ).

    Args:
        power (T): Total optical power [W].
        sigma_x (float): Gaussian standard deviation in x [m].
        sigma_y (float): Gaussian standard deviation in y [m].

    Returns:
        T: Peak (on-axis) intensity I₀ [W/m²].

    Example:
        >>> power_to_intensity_gaussian_beam(0.001, 0.5e-3, 0.5e-3)  # 1 mW, 0.5 mm sigma
        636.6197723675814

    Note:
        For circular beam with waist w₀, use sigma = w₀/2.
    """
    return power / (2 * np.pi * sigma_x * sigma_y)


def intensity_to_power_gaussian_beam(intensity: T, sigma_x: float, sigma_y: float) -> T:
    """Convert peak intensity to total power for Gaussian beam.

    Integrates Gaussian profile: P = I₀·2π·σₓ·σᵧ where I₀ is peak intensity.

    Args:
        intensity (T): Peak (on-axis) intensity I₀ [W/m²].
        sigma_x (float): Gaussian standard deviation in x [m].
        sigma_y (float): Gaussian standard deviation in y [m].

    Returns:
        T: Total optical power [W].

    Example:
        >>> intensity_to_power_gaussian_beam(1000.0, 0.5e-3, 0.5e-3)
        0.0015707963267948967
    """
    return intensity * 2 * np.pi * sigma_x * sigma_y


def power_to_rabi_gaussian_beam(
    power: T,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = ED_XtB,
) -> T:
    """Calculate peak Rabi frequency from power for Gaussian beam profile.

    Most common conversion for laser-molecule interactions. Gives on-axis (peak) Rabi
    frequency for Gaussian spatial profile.

    Args:
        power (T): Total optical power [W].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        sigma_x (float): Gaussian standard deviation in x [m].
        sigma_y (float): Gaussian standard deviation in y [m].
        D (float): Transition dipole moment [C·m]. Defaults to 2.6675506e-30 for
            TlF B-X optical transition at 272 nm.

    Returns:
        T: Peak (on-axis) Rabi frequency [rad/s].

    Example:
        >>> power_to_rabi_gaussian_beam(0.001, 1.0, 0.5e-3, 0.5e-3)  # 1 mW, TEM₀₀
        1.396...e+05

    Note:
        For circular beam with waist w₀, use sigma = w₀/2.
    """
    intensity = power_to_intensity_gaussian_beam(power, sigma_x, sigma_y)
    rabi = intensity_to_rabi(intensity, coupling, D)
    return rabi


def power_to_rabi_gaussian_beam_microwave(
    power: T,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = XConstants.D,
) -> T:
    """Calculate peak Rabi frequency from microwave power for Gaussian profile.

    Specialized for microwave transitions within ground X state. Uses larger permanent
    dipole moment appropriate for rotational transitions.

    Args:
        power (T): Total microwave power [W].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        sigma_x (float): Gaussian standard deviation in x [m].
        sigma_y (float): Gaussian standard deviation in y [m].
        D (float): Permanent dipole moment [C·m]. Defaults to 1.4103753e-29 for
            TlF X state (rotational transitions).

    Returns:
        T: Peak (on-axis) Rabi frequency [rad/s].

    Example:
        >>> power_to_rabi_gaussian_beam_microwave(0.1, 1.0, 5e-3, 5e-3)  # 100 mW
        7.377...e+05

    Note:
        Microwave dipole moment is ~5x larger than optical transition dipole.
    """
    return power_to_rabi_gaussian_beam(
        power=power, coupling=coupling, sigma_x=sigma_x, sigma_y=sigma_y, D=D
    )


def rabi_to_electric_field(rabi: T, coupling: float, D: float) -> T:
    """Calculate required electric field from desired Rabi frequency.

    Inverts the Rabi frequency formula: E₀ = Ω·ℏ/(μ·coupling).

    Args:
        rabi (T): Desired Rabi frequency [rad/s].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        D (float): Transition dipole moment [C·m].

    Returns:
        T: Required electric field amplitude [V/m].

    Example:
        >>> rabi_to_electric_field(1e6, 1.0, 2.67e-30)  # 1 MHz Rabi frequency
        3948.0...
    """
    return rabi * cst.hbar / (coupling * D)


def electric_field_to_intensity(electric_field: T) -> T:
    """Convert electric field amplitude to optical intensity.

    Uses time-averaged electromagnetic energy density: I = (1/2)·c·ε₀·E₀².

    Args:
        electric_field (T): Electric field amplitude E₀ [V/m].

    Returns:
        T: Optical intensity [W/m²].

    Example:
        >>> electric_field_to_intensity(868.36)  # Field of ~870 V/m
        999.99...
    """
    return 1 / 2 * cst.c * cst.epsilon_0 * electric_field**2


def rabi_to_power_gaussian_beam(
    rabi: T,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = ED_XtB,
) -> T:
    """Calculate required laser power from desired Rabi frequency for Gaussian beam.

    Inverse of power_to_rabi_gaussian_beam. Useful for determining laser power needed
    to achieve target Rabi frequency for specific transition.

    Args:
        rabi (T): Desired peak Rabi frequency [rad/s].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        sigma_x (float): Gaussian standard deviation in x [m].
        sigma_y (float): Gaussian standard deviation in y [m].
        D (float): Transition dipole moment [C·m]. Defaults to 2.6675506e-30 for
            TlF B-X optical transition.

    Returns:
        T: Required total optical power [W].

    Example:
        >>> rabi_to_power_gaussian_beam(1e6, 1.0, 0.5e-3, 0.5e-3)  # 1 MHz Rabi
        7.147...e-06
    """
    electric_field = rabi_to_electric_field(rabi, coupling, D)
    intensity = electric_field_to_intensity(electric_field)
    power = intensity_to_power_gaussian_beam(intensity, sigma_x, sigma_y)
    return power


def rabi_to_power_gaussian_beam_microwave(
    rabi: T,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = XConstants.D,
) -> T:
    """Calculate required microwave power from desired Rabi frequency for Gaussian beam.

    Specialized for microwave rotational transitions within ground X state.

    Args:
        rabi (T): Desired peak Rabi frequency [rad/s].
        coupling (float): Reduced matrix element coupling strength (dimensionless).
        sigma_x (float): Gaussian standard deviation in x [m].
        sigma_y (float): Gaussian standard deviation in y [m].
        D (float): Permanent dipole moment [C·m]. Defaults to 1.4103753e-29 for
            TlF X state rotational transitions.

    Returns:
        T: Required total microwave power [W].

    Example:
        >>> rabi_to_power_gaussian_beam_microwave(1e6, 1.0, 5e-3, 5e-3)
        1.356...e-07
    """
    return rabi_to_power_gaussian_beam(rabi, coupling, sigma_x, sigma_y, D)



from . import (
    B_coupled_Omega,
    B_uncoupled,
    X_uncoupled,
    basis_transformations,
    constants,
    generate_hamiltonian,
    matrix_elements_electric_dipole,
    matrix_elements_electric_quadrupole,
    quantum_operators,
    reduced_hamiltonian,
    utils,
    wigner,
)
from .basis_transformations import *  # noqa
from .constants import *  # noqa
from .generate_hamiltonian import *  # noqa
from .matrix_elements_electric_dipole import *  # noqa
from .matrix_elements_electric_quadrupole import *  # noqa
from .quantum_operators import *  # noqa
from .reduced_hamiltonian import *  # noqa
from .utils import *  # noqa
from .wigner import *  # noqa

__all__ = ["B_coupled_Omega", "X_uncoupled", "B_uncoupled"]
__all__ += generate_hamiltonian.__all__.copy()
__all__ += wigner.__all__.copy()
__all__ += constants.__all__.copy()
__all__ += quantum_operators.__all__.copy()
__all__ += utils.__all__.copy()
__all__ += basis_transformations.__all__.copy()
__all__ += reduced_hamiltonian.__all__.copy()
__all__ += matrix_elements_electric_dipole.__all__.copy()
__all__ += matrix_elements_electric_quadrupole.__all__.copy()



from pathlib import Path

import numpy as np

from centrex_tlf import couplings, states


def test_generate_coupling_matrix():
    qn_select = states.QuantumSelector(J=1)
    ground_states = states.generate_coupled_states_X(qn_select)

    qn_select = states.QuantumSelector(
        J=1, F1=1 / 2, F=1, P=1, Ω=1, electronic=states.ElectronicState.B
    )
    excited_states = states.generate_coupled_states_B(qn_select)

    QN = [1 * s for s in ground_states + excited_states]
    ground_states = [1 * s for s in ground_states]
    excited_states = [1 * s for s in excited_states]

    pol_vec = np.array([0.0, 0.0, 1.0])
    reduced = False
    normalize_pol = True

    coupling_matrix = couplings.generate_coupling_matrix(
        QN, ground_states, excited_states, pol_vec, reduced, normalize_pol
    )

    coupling_matrix_test = np.load(Path(__file__).parent / "coupling_matrix_test.npy")

    assert np.allclose(coupling_matrix, -coupling_matrix_test)


def test_generate_coupling_field_automatic():
    qn_select = states.QuantumSelector(J=1)
    ground_states = states.generate_coupled_states_X(qn_select)

    qn_select = states.QuantumSelector(
        J=1, F1=1 / 2, F=1, P=1, Ω=1, electronic=states.ElectronicState.B
    )
    excited_states = states.generate_coupled_states_B(qn_select)

    QN = [1 * s for s in ground_states + excited_states]
    ground_states = [1 * s for s in ground_states]
    excited_states = [1 * s for s in excited_states]

    H_rot = np.eye(len(QN), dtype=complex) * np.arange(len(QN))
    V_ref = np.eye(len(QN))
    pol_vecs = [np.array([0.0, 0.0, 1.0]), np.array([1.0, 0.0, 0.0])]
    normalize_pol = True

    coupling = couplings.generate_coupling_field_automatic(
        ground_states, excited_states, QN, H_rot, QN, V_ref, pol_vecs, normalize_pol
    )

    assert np.allclose(coupling.main_coupling, (-0.33333333333333326 + 0j))
    assert coupling.ground_main == ground_states[-3]
    assert coupling.excited_main == excited_states[1]
    assert np.all(coupling.fields[0].polarization == np.array([0.0, 0.0, 1.0]))
    assert np.all(coupling.fields[1].polarization == np.array([1.0, 0.0, 0.0]))



from . import (
    find_states,
    generate_states,
    population,
    states,
    utils,
    utils_compact,
)
from .find_states import *  # noqa
from .generate_states import *  # noqa
from .population import *  # noqa
from .states import *  # noqa
from .utils import *  # noqa
from .utils_compact import *  # noqa

__all__ = states.__all__.copy()
__all__ += utils.__all__.copy()
__all__ += utils_compact.__all__.copy()
__all__ += generate_states.__all__.copy()
__all__ += find_states.__all__.copy()
__all__ += population.__all__.copy()



import numpy as np
import pytest

from centrex_tlf import states


def test_state_initialization():
    s = states.CoupledBasisState(
        F=1, mF=0, F1=1 / 2, J=0, I1=1 / 2, I2=1 / 2, Omega=0, P=1
    )
    assert s == s
    assert isinstance(s, type(s))
    assert isinstance(s, states.CoupledBasisState)
    assert s.F == 1
    assert s.F1 == 1 / 2
    assert s.mF == 0
    assert s.J == 0
    assert s.I1 == 1 / 2
    assert s.I2 == 1 / 2
    assert s.Omega == 0
    assert s.P == 1


def test_state_matmul():
    QN_coupled = states.generate_coupled_states_ground(Js=[0, 1])
    QN_uncoupled = states.generate_uncoupled_states_ground(Js=[0, 1])
    assert QN_coupled[0] @ QN_coupled[0] == 1
    assert QN_coupled[0] @ QN_coupled[1] == 0
    assert QN_uncoupled[0] @ QN_uncoupled[0] == 1
    assert QN_uncoupled[0] @ QN_uncoupled[1] == 0
    assert QN_coupled[0] @ QN_uncoupled[1] == -0.7071067811865476 + 0j
    assert QN_coupled[0] @ QN_uncoupled[0] == 0 + 0j
    assert QN_uncoupled[1] @ QN_coupled[0] == -0.7071067811865476 + 0j
    assert QN_uncoupled[0] @ QN_coupled[0] == 0 + 0j
    with pytest.raises(TypeError):
        QN_coupled[0] @ 2


def test_basisstate_add():
    QN_coupled = states.generate_coupled_states_ground(Js=[0, 1])
    QN_uncoupled = states.generate_uncoupled_states_ground(Js=[0, 1])
    state1 = QN_coupled[0]
    state2 = QN_coupled[5]
    assert state1 + state2 == states.CoupledState([(1, state1), (1, state2)])
    with pytest.raises(TypeError):
        QN_coupled[0] + QN_uncoupled[5]


def test_CGc():
    assert states.CGc(1.0, -1.0, 1.0, 1.0, 1.0, 0.0) == -0.7071067811865475 + 0j


def test_parity_X():
    assert states.parity_X(0) == 1
    assert states.parity_X(1) == -1


def test_reorder_evecs():
    V_in = np.eye(10, dtype=np.complex128)
    E_in = np.random.rand(10).astype(np.complex128)
    V_ref = V_in.copy()
    ida = np.random.randint(0, 10)
    idb = np.random.randint(0, 10)
    _ = V_ref[ida, :].copy()
    V_ref[ida, :] = V_ref[idb, :]
    V_ref[idb, :] = _

    E_out, V_out = states.reorder_evecs(V_in, E_in, V_ref)
    assert E_out[ida] == E_in[idb]
    assert E_out[idb] == E_in[ida]


def test_quantumselector_get_indices():
    QN = states.generate_coupled_states_ground(Js=[0, 1])
    qn_select = states.QuantumSelector(J=1, mF=0, electronic=states.ElectronicState.X)
    assert np.all(
        qn_select.get_indices(QN, mode="python")
        == np.array([4, 6, 9, 13], dtype=np.int64)
    )
    assert np.all(
        qn_select.get_indices(QN, mode="julia")
        == np.array([4, 6, 9, 13], dtype=np.int64) + 1
    )


def test_find_closest_vector_idx():
    vector_array = np.eye(10).astype(np.complex128)
    idx_compare = np.random.randint(0, 10)
    state_vec = vector_array[idx_compare, :]
    idx = states.find_closest_vector_idx(state_vec, vector_array)
    assert idx == idx_compare


def test_check_approx_state_exact_state():
    QN = states.generate_coupled_states_ground(Js=[0, 1])
    ida = np.random.randint(0, len(QN))
    idb = np.random.randint(0, len(QN))
    approx = 1 * QN[ida]
    exact = 1 * QN[ida] + 0.1 * QN[idb]
    states.check_approx_state_exact_state(approx, exact)


def test_BasisStates_from_State():
    QN = states.generate_coupled_states_ground(Js=[0, 1])
    QN_State = 1 * QN
    QN_BasisState = states.BasisStates_from_State(QN_State)
    assert np.all(QN == QN_BasisState)


def test_get_indices_quantumnumbers_base():
    QN = states.generate_coupled_states_ground(Js=[0, 1])
    qn_select = states.QuantumSelector(J=1, mF=0, electronic=states.ElectronicState.X)
    indices = states.get_indices_quantumnumbers_base(qn_select, QN, mode="python")
    assert np.all(indices == np.array([4, 6, 9, 13], dtype=np.int64))
    indices = states.get_indices_quantumnumbers_base(qn_select, QN, mode="julia")
    assert np.all(indices == (np.array([4, 6, 9, 13], dtype=np.int64) + 1))


def test_get_indices_quantumnumbers():
    QN = states.generate_coupled_states_ground(Js=[0, 1])
    qn_select = [
        states.QuantumSelector(J=1, mF=0, electronic=states.ElectronicState.X),
        states.QuantumSelector(J=1, mF=1, electronic=states.ElectronicState.X),
    ]
    indices = states.get_indices_quantumnumbers(qn_select, QN)
    assert np.all(indices == np.array([4, 6, 7, 9, 10, 13, 14], dtype=np.int64))


def test_get_unique_basisstates():
    QN_original = states.generate_coupled_states_ground(Js=[0, 1])
    QN = np.append(QN_original, QN_original)
    QN_unique = states.get_unique_basisstates_from_basisstates(QN)
    assert np.all(QN_original == QN_unique)


def test_state_add():
    ss = states.generate_coupled_states_ground(Js=[0, 1])
    state1 = states.CoupledState([(1.0, s) for s in ss])
    ss = states.generate_coupled_states_ground(Js=[1, 2])
    state2 = states.CoupledState([(2.0, s) for s in ss])
    state_sum = state1 + state2
    state_sum = state_sum.order_by_amp()
    for a, s in state_sum.data:
        if s.J == 0:
            assert a == 1.0
        elif s.J == 1:
            assert a == 3.0
        elif s.J == 2:
            assert a == 2.0


def test_hash():
    # Test that all states have unique hashes (no collisions)
    s = states.generate_uncoupled_states_ground(np.arange(0, 15))
    assert (
        len([si.__hash__() for si in s]) == np.unique([si.__hash__() for si in s]).size
    )
    s = states.generate_coupled_states_ground(np.arange(0, 15))
    assert (
        len([si.__hash__() for si in s]) == np.unique([si.__hash__() for si in s]).size
    )

    s = states.generate_coupled_states_excited(
        np.arange(0, 15),
        Ps=None,
        Omegas=[-1, 1],  # type: ignore[arg-type]
    )
    assert (
        len([si.__hash__() for si in s]) == np.unique([si.__hash__() for si in s]).size
    )


def test_hash_stability_and_uniqueness():
    """Test that hash values are stable and unique for both CoupledBasisState and UncoupledBasisState.
    
    This test verifies:
    1. Hash values are consistent across multiple calls
    2. Equal states produce equal hashes
    3. Different states produce different hashes
    4. No collisions occur in large state spaces
    """
    # Test CoupledBasisState hash stability
    state1 = states.CoupledBasisState(
        F=1, mF=0, F1=1/2, J=0, I1=1/2, I2=1/2, Omega=0, P=1
    )
    hash1_call1 = hash(state1)
    hash1_call2 = hash(state1)
    assert hash1_call1 == hash1_call2, "Hash should be stable across calls"
    
    # Test equal states have equal hashes
    state1_copy = states.CoupledBasisState(
        F=1, mF=0, F1=1/2, J=0, I1=1/2, I2=1/2, Omega=0, P=1
    )
    assert hash(state1) == hash(state1_copy), "Equal states must have equal hashes"
    assert state1 == state1_copy, "States should be equal"
    
    # Test different states have different hashes
    state2 = states.CoupledBasisState(
        F=1, mF=1, F1=1/2, J=0, I1=1/2, I2=1/2, Omega=0, P=1
    )
    assert hash(state1) != hash(state2), "Different states should have different hashes"
    assert state1 != state2, "States should not be equal"
    
    # Test UncoupledBasisState hash stability
    ustate1 = states.UncoupledBasisState(
        J=0, mJ=0, I1=1/2, m1=1/2, I2=1/2, m2=-1/2, P=1, Omega=0,
        electronic_state=states.ElectronicState.X
    )
    uhash1_call1 = hash(ustate1)
    uhash1_call2 = hash(ustate1)
    assert uhash1_call1 == uhash1_call2, "Hash should be stable across calls"
    
    # Test equal uncoupled states have equal hashes
    ustate1_copy = states.UncoupledBasisState(
        J=0, mJ=0, I1=1/2, m1=1/2, I2=1/2, m2=-1/2, P=1, Omega=0,
        electronic_state=states.ElectronicState.X
    )
    assert hash(ustate1) == hash(ustate1_copy), "Equal states must have equal hashes"
    assert ustate1 == ustate1_copy, "States should be equal"
    
    # Test different uncoupled states have different hashes
    ustate2 = states.UncoupledBasisState(
        J=0, mJ=0, I1=1/2, m1=-1/2, I2=1/2, m2=-1/2, P=1, Omega=0,
        electronic_state=states.ElectronicState.X
    )
    assert hash(ustate1) != hash(ustate2), "Different states should have different hashes"
    assert ustate1 != ustate2, "States should not be equal"
    
    # Test no collisions in large state space for coupled states
    coupled_ground = states.generate_coupled_states_ground(np.arange(0, 15))
    coupled_hashes = [hash(s) for s in coupled_ground]
    assert len(coupled_hashes) == len(set(coupled_hashes)), (
        f"Found hash collisions in coupled ground states: "
        f"{len(coupled_hashes)} states, {len(set(coupled_hashes))} unique hashes"
    )
    
    # Test no collisions in large state space for uncoupled states
    uncoupled_ground = states.generate_uncoupled_states_ground(np.arange(0, 15))
    uncoupled_hashes = [hash(s) for s in uncoupled_ground]
    assert len(uncoupled_hashes) == len(set(uncoupled_hashes)), (
        f"Found hash collisions in uncoupled ground states: "
        f"{len(uncoupled_hashes)} states, {len(set(uncoupled_hashes))} unique hashes"
    )
    
    # Test no collisions in excited states with various quantum numbers
    coupled_excited = states.generate_coupled_states_excited(
        np.arange(0, 10), Ps=None, Omegas=[-1, 1]  # type: ignore[arg-type]
    )
    excited_hashes = [hash(s) for s in coupled_excited]
    assert len(excited_hashes) == len(set(excited_hashes)), (
        f"Found hash collisions in coupled excited states: "
        f"{len(excited_hashes)} states, {len(set(excited_hashes))} unique hashes"
    )
    
    # Test that states can be used in sets (requires proper __hash__ and __eq__)
    state_set = {state1, state1_copy, state2}
    assert len(state_set) == 2, "Set should contain only unique states"
    assert state1 in state_set
    assert state2 in state_set
    
    # Test that states can be used as dictionary keys
    state_dict = {state1: "first", state2: "second"}
    assert state_dict[state1_copy] == "first", "Equal states should access same dict entry"
    assert state_dict[state2] == "second"



"""Transition selection and organization for optical Bloch equations.

This module provides tools for defining and organizing optical and microwave transitions
including their polarizations, symbolically-defined Rabi frequencies and detunings,
and the quantum states involved.
"""
from dataclasses import dataclass
from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
import sympy as smp

from centrex_tlf import states
from centrex_tlf.transitions import (
    MicrowaveTransition,
    OpticalTransition,
    OpticalTransitionType,
)

from .polarization import Polarization
from .utils import check_transition_coupled_allowed, select_main_states

__all__ = [
    "TransitionSelector",
    "generate_transition_selectors",
    "get_possible_optical_transitions",
]


def _sanitize_polarization_name(name: str) -> str:
    """Sanitize polarization name for use as a symbolic variable.
    
    Only keeps letters (a-z, A-Z). If the result is empty or contains
    mathematical symbols/numbers, returns None to trigger fallback naming.
    
    Args:
        name: Polarization name (e.g., "X", "1/√2 X + 1/√2 Z")
        
    Returns:
        Sanitized name containing only letters, or empty string if not simple
        
    Examples:
        >>> _sanitize_polarization_name("X")
        'X'
        >>> _sanitize_polarization_name("σp")
        'σp'
        >>> _sanitize_polarization_name("1/√2 X + 1/√2 Z")
        ''
    """
    # Check if name is simple (only letters and maybe σ, Ω, etc.)
    # Allow Greek letters but not math operators or numbers
    simple_chars = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
                      "αβγδεζηθικλμνξοπρστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ"
                      "σΣ")
    
    # If name contains operators, spaces, numbers, or other complex chars, return empty
    if any(c in name for c in "+-*/^√() 0123456789"):
        return ""
    
    # Keep only allowed characters
    sanitized = "".join(c for c in name if c in simple_chars)
    return sanitized


@dataclass
class TransitionSelector:
    """Describes a laser-driven transition with associated parameters.
    
    Attributes:
        ground (Sequence[CoupledState]): Ground states involved in the transition
        excited (Sequence[CoupledState]): Excited states involved in the transition
        polarizations (Sequence[npt.NDArray[np.complex128]]): Polarization vectors for
            each field component
        polarization_symbols (List[smp.Symbol]): Symbolic variables for polarization
            amplitudes
        Ω (smp.Symbol): Symbolic Rabi frequency
        δ (smp.Symbol): Symbolic detuning from resonance
        description (str | None): Human-readable description of the transition
        type (str | None): Type of transition (e.g., "optical", "microwave")
        ground_main (CoupledState | None): Main ground state for the transition
        excited_main (CoupledState | None): Main excited state for the transition
        phase_modulation (bool): Whether phase modulation is applied. Defaults to False.
    """
    ground: Sequence[states.CoupledState]
    excited: Sequence[states.CoupledState]
    polarizations: Sequence[npt.NDArray[np.complex128]]
    polarization_symbols: List[smp.Symbol]
    Ω: smp.Symbol
    δ: smp.Symbol
    description: Optional[str] = None
    type: Optional[str] = None
    ground_main: Optional[states.CoupledState] = None
    excited_main: Optional[states.CoupledState] = None
    phase_modulation: bool = False

    def __repr__(self) -> str:
        if self.description is None:
            J_g = np.unique([g.largest.J for g in self.ground])[0]
            J_e = np.unique([e.largest.J for e in self.excited])[0]
            return f"TransitionSelector(J={J_g} -> J={J_e})"
        else:
            return f"TransitionSelector({self.description})"


def generate_transition_selectors(
    transitions: Sequence[Union[OpticalTransition, MicrowaveTransition]],
    polarizations: Sequence[Sequence[Polarization]],
    ground_mains: Optional[Sequence[states.CoupledState]] = None,
    excited_mains: Optional[Sequence[states.CoupledState]] = None,
    phase_modulations: Optional[Sequence[bool]] = None,
) -> List[TransitionSelector]:
    """Generate TransitionSelector objects from transition and polarization specs.
    
    Creates a list of TransitionSelector objects that describe the laser-driven
    transitions in an optical Bloch equation system. Each selector includes the
    ground and excited states, polarization vectors, and symbolic parameters (Rabi
    frequency Ω and detuning δ) for use in symbolic calculations.

    Args:
        transitions (Sequence[OpticalTransition | MicrowaveTransition]): Transitions
            to include in the system
        polarizations (Sequence[Sequence[Polarization]]): List of polarizations for
            each transition. Inner sequence contains polarizations for a single
            transition (e.g., [pol_x, pol_y] for two-polarization transition).
        ground_mains (Sequence[CoupledState] | None): Optional main ground state for
            each transition. Used to identify the primary coupling. Defaults to None.
        excited_mains (Sequence[CoupledState] | None): Optional main excited state for
            each transition. Used to identify the primary coupling. Defaults to None.
        phase_modulations (Sequence[bool] | None): Whether each transition has phase
            modulation applied. Defaults to None (no modulation).

    Returns:
        List[TransitionSelector]: List of TransitionSelector objects, one per transition
        
    Example:
        >>> transitions = [OpticalTransition(J_ground=0, J_excited=1)]
        >>> pols = [[polarization_X, polarization_Y]]
        >>> selectors = generate_transition_selectors(transitions, pols)
    """
    transition_selectors = []

    for idt, (transition, polarization) in enumerate(zip(transitions, polarizations)):
        ground_states_approx_qn_select = states.QuantumSelector(
            J=transition.J_ground,
            electronic=transition.electronic_ground,
            P=transition.P_ground,
            Ω=transition.Ω_ground,
        )
        ground_states_approx = list(
            [
                1 * s
                for s in states.generate_coupled_states_X(
                    ground_states_approx_qn_select
                )
            ]
        )

        if isinstance(transition, OpticalTransition):
            excited_states_approx = [
                1 * s
                for s in states.generate_coupled_states_B(transition.qn_select_excited)
            ]
        elif isinstance(transition, MicrowaveTransition):
            excited_states_approx = [
                1 * s
                for s in states.generate_coupled_states_X(transition.qn_select_excited)
            ]
        else:
            raise TypeError(
                f"Transition must be OpticalTransition or MicrowaveTransition, "
                f"got {type(transition).__name__}"
            )

        if phase_modulations is None:
            phase_modulation = False
        else:
            phase_modulation = phase_modulations[idt]

        # Initialize main states
        ground_main: Optional[states.CoupledState] = None
        excited_main: Optional[states.CoupledState] = None
        if ground_mains is None:
            ground_main, excited_main = select_main_states(
                ground_states_approx, excited_states_approx, polarization[0].vector
            )

        # Create polarization symbols with sanitized names
        pol_symbols = []
        for idx, p in enumerate(polarization):
            sanitized = _sanitize_polarization_name(p.name)
            if sanitized:
                # Use sanitized name if it's simple (e.g., "X", "Y", "σp")
                pol_symbols.append(smp.Symbol(f"P{sanitized}{idt}"))
            else:
                # Use generic naming for complex expressions (e.g., "PA0", "PB0")
                # A for first polarization, B for second, etc.
                pol_label = chr(ord('A') + idx)
                pol_symbols.append(smp.Symbol(f"P{pol_label}{idt}"))

        transition_selectors.append(
            TransitionSelector(
                ground=ground_states_approx,
                excited=excited_states_approx,
                polarizations=[p.vector for p in polarization],
                polarization_symbols=pol_symbols,
                Ω=smp.Symbol(f"Ω{idt}", complex=True),
                δ=smp.Symbol(f"δ{idt}"),
                description=transition.name,
                ground_main=ground_main if ground_mains is None else ground_mains[idt],
                excited_main=excited_main
                if excited_mains is None
                else excited_mains[idt],
                phase_modulation=phase_modulation,
            )
        )
    return transition_selectors


def get_possible_optical_transitions(
    ground_state: states.CoupledBasisState,
    transition_types: Optional[Sequence[OpticalTransitionType]] = None,
):
    J = ground_state.J
    # F1 = ground_state.F1
    # F = ground_state.F
    I1 = float(ground_state.I1)
    I2 = float(ground_state.I2)

    if transition_types is None:
        transition_types = [t for t in OpticalTransitionType]

    transitions = []
    for transition_type in transition_types:
        ΔJ = transition_type.value
        J_excited = J + ΔJ
        _transitions = [
            OpticalTransition(transition_type, J, float(F1), int(F))
            for F1 in np.arange(np.abs(J_excited - I1), J_excited + I1 + 1)
            for F in np.arange(np.abs(F1 - I2), F1 + I2 + 1, dtype=int)
        ]
        _transitions = [
            t
            for t in _transitions
            if check_transition_coupled_allowed(ground_state, t.excited_states[0])
        ]
        transitions.append(_transitions)
    return transitions



import numpy as np

from centrex_tlf import hamiltonian, transitions

trans = transitions.OpticalTransition(
    transitions.OpticalTransitionType.R, J_ground=0, F1_excited=3 / 2, F_excited=2
)

# include Jmax_X to include mixing of ground states. The B state calculation
# automatically uses Jmax_B = the largest J included + 2
H_red_electric = hamiltonian.generate_reduced_hamiltonian_transitions(
    [trans], E=np.array([0, 0, 500]), Jmax_X=4
)



import numpy as np

from centrex_tlf import couplings, states


def test_calculate_br():
    excited_state = states.CoupledBasisState(
        J=1,
        F1=1 / 2,
        F=1,
        mF=0,
        I1=1 / 2,
        I2=1 / 2,
        Omega=1,
        P=1,
        electronic_state=states.ElectronicState.B,
    )
    qn_select = states.QuantumSelector(J=1)
    ground_states = states.generate_coupled_states_X(qn_select)
    br = couplings.calculate_br(1 * excited_state, [1 * s for s in ground_states])
    assert np.allclose(
        br,
        [
            0.22222222,
            0.22222222,
            0.0,
            0.22222222,
            0.02777778,
            0.0,
            0.02777778,
            0.0,
            0.08333333,
            0.11111111,
            0.08333333,
            0.0,
        ],
    )
    assert br.sum() == 1.0



from typing import Sequence, Union

import numpy as np
import numpy.typing as npt

from centrex_tlf.states import BasisState

__all__ = ["generate_transform_matrix"]


def generate_transform_matrix(
    basis1: Union[Sequence[BasisState], npt.NDArray],
    basis2: Union[Sequence[BasisState], npt.NDArray],
) -> npt.NDArray[np.complex128]:
    """Generate transformation matrix between two quantum state bases.
    
    Computes the transformation matrix S that converts operators from basis1 to basis2:
        H₂ = S† @ H₁ @ S
    
    where S[i,j] = ⟨basis1[i]|basis2[j]⟩ is the overlap matrix between the two bases.
    
    Args:
        basis1 (Sequence[BasisState] | npt.NDArray): Sequence or array of basis states 
            defining the first basis
        basis2 (Sequence[BasisState] | npt.NDArray): Sequence or array of basis states 
            defining the second basis
    
    Returns:
        npt.NDArray[np.complex128]: Transformation matrix S of shape (n, n) that 
            converts operators from basis1 to basis2
    
    Raises:
        ValueError: If the two bases have different dimensions or if either basis is empty
    
    Example:
        >>> # Transform uncoupled to coupled basis
        >>> S = generate_transform_matrix(uncoupled_basis, coupled_basis)
        >>> H_coupled = S.conj().T @ H_uncoupled @ S
    
    Note:
        For large bases (n > 100), consider using numba JIT compilation or 
        vectorized operations for better performance.
    """
    # Input validation
    n1 = len(basis1)
    n2 = len(basis2)
    
    if n1 == 0 or n2 == 0:
        raise ValueError("Both bases must be non-empty")
    
    if n1 != n2:
        raise ValueError(
            f"Bases must have the same dimension: basis1 has {n1} states, "
            f"basis2 has {n2} states"
        )
    
    # Initialize transformation matrix
    n = n1
    S = np.zeros((n, n), dtype=np.complex128)

    # Calculate inner products: S[i,j] = ⟨basis1[i]|basis2[j]⟩
    for i, state1 in enumerate(basis1):
        for j, state2 in enumerate(basis2):
            S[i, j] = state1 @ state2

    return S



import copy
from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
from scipy import constants as cst

from .find_states import QuantumSelector
from .states import CoupledState
from .utils_compact import compact_QN_coupled_indices

__all__ = [
    "thermal_population",
    "J_levels",
    "generate_thermal_population_states",
    "generate_population_states",
]


def thermal_population(J: int, T: float, B: float = 6.66733e9, n: int = 100) -> float:
    """Calculate the thermal population of a given J sublevel.

    Uses Boltzmann distribution to calculate the relative population of a rotational
    sublevel at a given temperature.

    Args:
        J (int): Rotational level
        T (float): Temperature in Kelvin
        B (float, optional): Rotational constant in Hz. Defaults to 6.66733e9 (TlF ground state)
        n (int, optional): Number of rotational levels to include in partition function normalization.
                          Defaults to 100.

    Returns:
        float: Relative population in the rotational sublevel (normalized by partition function)

    Raises:
        ValueError: If T is non-positive
    """
    if T <= 0:
        raise ValueError(f"Temperature must be positive, got {T} K")

    c = 2 * np.pi * cst.hbar * B / (cst.k * T)

    def energy_factor(J_level: int) -> float:
        """Calculate the Boltzmann factor for a given J level."""
        return -c * J_level * (J_level + 1)

    # Partition function
    Z = np.sum([J_levels(i) * np.exp(energy_factor(i)) for i in range(n)])
    return J_levels(J) * np.exp(energy_factor(J)) / Z


def J_levels(J: int) -> int:
    """Calculate the number of hyperfine sublevels per J rotational level.

    For TlF, each J level has 4*(2J+1) hyperfine sublevels due to the two nuclear
    spins (I1=1/2 for Tl, I2=1/2 for F).

    Args:
        J (int): Rotational quantum number

    Returns:
        int: Number of hyperfine sublevels for the given J
    """
    return 4 * (2 * J + 1)


def generate_thermal_population_states(
    states_to_fill: Union[Sequence[QuantumSelector], QuantumSelector],
    states: Sequence[CoupledState],
    T: float,
    qn_compact: Optional[Union[Sequence[QuantumSelector], QuantumSelector]] = None,
) -> npt.NDArray[np.complex128]:
    """Generate a thermal distribution over specified states.

    Creates a density matrix with thermal (Boltzmann) population distribution over
    the states specified by states_to_fill. Population is distributed uniformly within
    each J manifold according to the thermal population of that J level.

    Args:
        states_to_fill (Union[Sequence[QuantumSelector], QuantumSelector]): QuantumSelector
            or sequence of QuantumSelectors specifying which states to populate thermally
        states (Sequence[CoupledState]): All states used in the simulation
        T (float): Temperature in Kelvin
        qn_compact (Optional[Union[Sequence[QuantumSelector], QuantumSelector]], optional):
            QuantumSelector(s) specifying states to compact in the OBE system. Population
            in compacted states is summed together. Defaults to None.

    Returns:
        npt.NDArray[np.complex128]: Density matrix with trace normalized to 1

    Raises:
        ValueError: If states_to_fill doesn't have J levels defined
        TypeError: If states_to_fill is not a QuantumSelector or sequence of them
    """
    # branch for single QuantumSelector use
    if isinstance(states_to_fill, QuantumSelector):
        # get all involved Js
        Js = states_to_fill.J
        if Js is None:
            raise ValueError(
                "states_to_fill needs rotational levels (J) defined to generate "
                "thermal population density"
            )
        # check if J was a list
        _Js = (
            np.array([Js])
            if not isinstance(Js, (np.ndarray, list, tuple, Sequence))
            else np.array(Js)
        )
        # get indices of states to fill
        indices_to_fill = states_to_fill.get_indices(states)
    # branch for multiple QuantumSelectors use
    elif isinstance(states_to_fill, (list, np.ndarray, tuple)):
        if not isinstance(states_to_fill[0], QuantumSelector):
            raise TypeError(
                f"Need to supply a sequence of QuantumSelectors, not {type(states_to_fill[0])}"
            )
        # get all involved Js
        _Js = np.array([], dtype=np.int_)
        for stf in states_to_fill:
            J = stf.J
            # check if J was a list
            if J is not None:
                _J = (
                    [J] if not isinstance(J, (np.ndarray, list, tuple, Sequence)) else J
                )
                _Js = np.append(_Js, _J)
        if len(_Js) == 0:
            raise ValueError(
                "states_to_fill requires QuantumSelectors with rotational levels (J) defined"
            )
        # get indices of states to fill
        indices_to_fill = np.array([], dtype=np.int_)
        for stf in states_to_fill:
            indices_to_fill = np.append(
                indices_to_fill, stf.get_indices(states)
            ).astype(int)
    else:
        raise TypeError(
            "states_to_fill required to be a QuantumsSelector or a Sequence of "
            f"QuantumSelectors, not {type(states_to_fill)}"
        )

    # remove duplicates from Js and indices_to_fill
    _Js = np.unique(_Js)
    indices_to_fill = np.unique(indices_to_fill)

    # thermal population per hyperfine level for each involved J
    thermal_populations = dict(
        [(Ji, thermal_population(Ji, T) / J_levels(Ji)) for Ji in _Js]
    )
    # generate an empty density matrix
    density = np.zeros(len(states), dtype=complex)
    # fill the density matrix
    for idd in indices_to_fill:
        state = states[idd].largest
        thermal_pop = thermal_populations[state.J]
        density[idd] = thermal_pop

    if qn_compact is not None:
        states_compact = copy.copy(states)
        if isinstance(qn_compact, (list, tuple, np.ndarray, Sequence)):
            for qnc in qn_compact:
                indices_compact = qnc.get_indices(states_compact)
                pop_compact = density[indices_compact].sum()
                density = np.array(
                    [di for idd, di in enumerate(density) if idd not in indices_compact]
                )
                density[indices_compact[0]] = pop_compact
                states_compact = compact_QN_coupled_indices(
                    states_compact, indices_compact
                )
        elif isinstance(qn_compact, QuantumSelector):
            indices_compact = qn_compact.get_indices(states_compact)
            pop_compact = density[indices_compact].sum()
            density = np.array(
                [di for idd, di in enumerate(density) if idd not in indices_compact]
            )
            density[indices_compact[0]] = pop_compact
            states_compact = compact_QN_coupled_indices(states_compact, indices_compact)
        else:
            raise TypeError(
                f"qn_compact must be a QuantumSelector or a sequence of QuantumSelectors, "
                f"not {type(qn_compact)}"
            )

    density = np.eye(len(density), dtype=np.complex128) * density

    # normalize the trace to 1 and return the density matrix
    return density / np.trace(density)


def generate_population_states(
    states: Union[List[int], npt.NDArray[np.int_]], levels: int
) -> npt.NDArray[np.complex128]:
    """Generate a uniform population distribution over specified states.

    Creates a density matrix with equal population in each of the specified states,
    normalized so the trace equals 1.

    Args:
        states (Union[List[int], npt.NDArray[np.int_]]): Indices of states to populate
        levels (int): Total number of levels in the system

    Returns:
        npt.NDArray[np.complex128]: Density matrix with uniform population in specified
            states, trace normalized to 1

    Raises:
        ValueError: If any state index is out of bounds
    """
    states_array = np.asarray(states)
    if np.any(states_array < 0) or np.any(states_array >= levels):
        raise ValueError(
            f"State indices must be in range [0, {levels}), got indices: {states_array}"
        )

    density = np.zeros([levels, levels], dtype=np.complex128)
    for state in states_array:
        density[state, state] = 1
    return density / np.trace(density)



from centrex_tlf import hamiltonian, states

# generate the hyperfine sublevels in J=0 and J=1
QN = states.generate_uncoupled_states_ground(Js=[0, 1])
# generate the coupled hyperfine sublevels in J=0 and J=1
QNc = states.generate_coupled_states_ground(Js=[0, 1])

# generate a dictionary with X hamiltonian terms
H = hamiltonian.generate_uncoupled_hamiltonian_X(QN)
Hfunc = hamiltonian.generate_uncoupled_hamiltonian_X_function(H)
H0 = Hfunc(E=[0, 0, 0], B=[0, 0, 1e-3])

# generate the transformation matrix
transform = hamiltonian.generate_transform_matrix(QN, QNc)

# calculate the transformed matrix
H0c = transform.conj().T @ H0 @ transform



import numpy as np
import sympy as smp

from centrex_tlf import lindblad, states


def get_insert_indices():
    X_states = states.generate_coupled_states_ground(Js=[0, 1, 2])
    B_states = states.generate_coupled_states_excited(Js=[1], Ps=-1, Omegas=1)
    qn_select_excited = states.QuantumSelector(J=1, electronic=states.ElectronicState.B)
    decay_channels = [
        lindblad.DecayChannel(
            1 * states.CoupledBasisState(None, None, None, None, None, None, v="other"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching",
        ),
    ]
    indices = lindblad.utils_decay.get_insert_level_indices(
        decay_channels, list(X_states) + list(B_states), B_states
    )
    assert type(indices) == list
    assert indices == [36]


def test_add_level_symbolic_hamiltonian():
    original = smp.ones(4)
    new = lindblad.utils_decay.add_level_symbolic_hamiltonian(original, 2)
    assert new.shape == (5, 5)
    assert new[:, 2] == smp.zeros(5, 1)
    assert new[2, :] == smp.zeros(1, 5)


def test_add_levels_symbolic_hamiltonian():
    X_states = states.generate_coupled_states_ground(Js=[0, 1, 2])
    B_states = states.generate_coupled_states_excited(Js=[1], Ps=-1, Omegas=1)
    qn_select_excited = states.QuantumSelector(J=1, electronic=states.ElectronicState.B)

    size = len(X_states) + len(B_states)

    H = smp.ones(size)
    decay_channels = [
        lindblad.DecayChannel(
            1 * states.CoupledBasisState(None, None, None, None, None, None, v="other"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching",
        ),
        lindblad.DecayChannel(
            1
            * states.CoupledBasisState(None, None, None, None, None, None, v="other 1"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching",
        ),
    ]
    indices, arr = lindblad.utils_decay.add_levels_symbolic_hamiltonian(
        H, decay_channels, list(X_states) + list(B_states), B_states
    )
    assert indices == [36, 37]
    assert arr[:, 36] == smp.zeros(size + len(decay_channels), 1)
    assert arr[36, :] == smp.zeros(1, size + len(decay_channels))
    assert arr[:, 37] == smp.zeros(size + len(decay_channels), 1)
    assert arr[37, :] == smp.zeros(1, size + len(decay_channels))


def test_add_states_QN():
    X_states = states.generate_coupled_states_ground(Js=[0, 1, 2])
    B_states = states.generate_coupled_states_excited(Js=[1], Ps=-1, Omegas=1)
    qn_select_excited = states.QuantumSelector(J=1, electronic=states.ElectronicState.B)

    size = len(X_states) + len(B_states)

    decay_channels = [
        lindblad.DecayChannel(
            1 * states.CoupledBasisState(None, None, None, None, None, None, v="other"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching",
        ),
        lindblad.DecayChannel(
            1
            * states.CoupledBasisState(None, None, None, None, None, None, v="other 1"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching",
        ),
    ]

    indices = lindblad.utils_decay.get_insert_level_indices(
        decay_channels, list(X_states) + list(B_states), B_states
    )

    QN_new = lindblad.utils_decay.add_states_QN(
        decay_channels, list(X_states) + list(B_states), indices
    )

    assert len(QN_new) == size + len(decay_channels)
    assert QN_new[36] == decay_channels[0].ground
    assert QN_new[37] == decay_channels[1].ground


def test_add_decays_C_arrays():
    X_states = states.generate_coupled_states_ground(Js=[0, 1, 2])
    B_states = states.generate_coupled_states_excited(Js=[1], Ps=-1, Omegas=1)
    qn_select_excited = states.QuantumSelector(J=1, electronic=states.ElectronicState.B)

    size = len(X_states) + len(B_states)

    decay_channels = [
        lindblad.DecayChannel(
            1 * states.CoupledBasisState(None, None, None, None, None, None, v="other"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching",
        ),
        lindblad.DecayChannel(
            1
            * states.CoupledBasisState(None, None, None, None, None, None, v="other 1"),
            qn_select_excited,
            branching=1e-2,
            description="vibrational branching 1",
        ),
    ]
    indices = lindblad.utils_decay.get_insert_level_indices(
        decay_channels, list(X_states) + list(B_states), B_states
    )

    C_array = np.zeros((3, size, size))
    C_array[0, -1, 0] = 1
    C_array[1, -2, 0] = 1
    C_array[2, -3, 0] = 1
    C_array_new = lindblad.utils_decay.add_decays_C_arrays(
        decay_channels, indices, list(X_states) + list(B_states), C_array, Γ=1
    )
    assert C_array_new.shape == (27, 50, 50)
    assert C_array_new.sum() == 5.4 + 0j



"""Utilities for compacting quantum state manifolds in coupling calculations.

This module provides functions for reducing the dimensionality of coupling matrices
and branching ratio arrays by combining degenerate or nearly-degenerate states into
single manifolds.
"""
import copy
from typing import List, Sequence, TypeVar, Union

import numpy as np
import numpy.typing as npt

from centrex_tlf import states

from .coupling_matrix import CouplingField, CouplingFields

__all__: List[str] = [
    "delete_row_column_indices",
    "insert_row_column_indices",
    "BR_to_C_array",
    "C_array_to_BR",
    "compact_BR_array_indices",
    "compact_C_array_indices",
    "compact_coupling_field",
    "compact_coupling_field_indices",
]


DType = TypeVar("DType", bound=np.generic)


def delete_row_column_indices(
    arr: npt.NDArray[DType], indices_compact: Union[Sequence[int], npt.NDArray[np.int_]]
) -> npt.NDArray[DType]:
    """Delete specified rows and columns from a 2D array, excluding the first index.
    
    Removes rows and columns at indices_compact[1:] from the array, preserving
    indices_compact[0]. Used when compacting states where the first index becomes
    the representative.
    
    Args:
        arr (npt.NDArray[DType]): 2D array to modify
        indices_compact (Sequence[int] | npt.NDArray[np.int_]): Indices to remove
            (all except first)
    
    Returns:
        npt.NDArray[DType]: Array with specified rows and columns removed
    """
    return np.delete(
        np.delete(arr, indices_compact[1:], axis=0), indices_compact[1:], axis=1
    )


def insert_row_column_indices(
    arr: npt.NDArray[DType], indices_insert: Union[Sequence[int], npt.NDArray[np.int_]]
) -> npt.NDArray[DType]:
    """Insert zero-filled rows and columns at specified indices.
    
    Expands a 2D array by inserting zero rows and columns at the specified indices.
    Inverse operation of delete_row_column_indices.
    
    Args:
        arr (npt.NDArray[DType]): 2D array to expand
        indices_insert (Sequence[int] | npt.NDArray[np.int_]): Indices where to insert
            zeros
    
    Returns:
        npt.NDArray[DType]: Expanded array with zeros at specified locations
    """
    arr = arr.copy()
    insert = np.zeros(arr.shape[1], dtype=arr.dtype)
    for idi in indices_insert:
        arr = np.insert(arr, idi, insert, axis=0)
    insert = np.zeros(arr.shape[0], dtype=arr.dtype)
    for idi in indices_insert:
        arr = np.insert(arr, idi, insert, axis=1)
    return arr


def BR_to_C_array(BR: npt.NDArray[np.floating], Gamma: float) -> npt.NDArray[np.floating]:
    """Convert branching ratios to collapse operator matrix elements.
    
    Computes C = √(BR * Γ) for Lindblad collapse operators.
    
    Args:
        BR (npt.NDArray[np.floating]): Branching ratio array
        Gamma (float): Decay rate in rad/s
    
    Returns:
        npt.NDArray[np.floating]: Collapse operator array
    """
    return np.sqrt(BR * Gamma)


def C_array_to_BR(
    C_array: npt.NDArray[np.floating], Gamma: float
) -> npt.NDArray[np.floating]:
    """Convert collapse operator matrix elements to branching ratios.
    
    Computes BR = C² / Γ, inverse of BR_to_C_array.
    
    Args:
        C_array (npt.NDArray[np.floating]): Collapse operator array
        Gamma (float): Decay rate in rad/s
    
    Returns:
        npt.NDArray[np.floating]: Branching ratio array
    """
    return C_array**2 / Gamma


def compact_BR_array_indices(
    BR_array: npt.NDArray[np.floating], indices_compact: npt.NDArray[np.int_]
) -> npt.NDArray[np.floating]:
    new_shape = np.asarray(BR_array[0].shape) - len(indices_compact) + 1
    BR_array_new = []
    for BR in BR_array:
        id1, id2 = np.nonzero(BR)
        id1 = id1[0]
        id2 = id2[0]
        if (id1 not in indices_compact) & (id2 not in indices_compact):
            BR_array_new.append(delete_row_column_indices(BR, indices_compact))

    BR_sum = np.zeros(new_shape, "complex")
    BR_sum[indices_compact[0], :] = np.sum(BR_array, axis=0)[indices_compact].sum(
        axis=0
    )[
        -new_shape[0] :  # noqa: 203
    ]
    for id1, id2 in zip(*np.nonzero(BR_sum)):
        BR_new = np.zeros(new_shape, "complex")
        BR_new[id1, id2] = BR_sum[id1, id2]
        BR_array_new.append(BR_new)
    return np.array(BR_array_new)


def compact_C_array_indices(
    C_array: npt.NDArray[np.floating],
    Gamma: float,
    indices_compact: npt.NDArray[np.int_],
) -> npt.NDArray[np.floating]:
    BR = C_array_to_BR(C_array, Gamma)
    BR_compact = compact_BR_array_indices(BR, indices_compact)
    C_array_compact = BR_to_C_array(BR_compact, Gamma)
    return C_array_compact


def compact_coupling_field(
    fields: CouplingFields,
    QN: Sequence[states.CoupledState],
    qn_compact: Union[Sequence[states.QuantumSelector], states.QuantumSelector],
) -> CouplingFields:
    if isinstance(qn_compact, states.QuantumSelector):
        qn_compact = [qn_compact]

    QN_compact = copy.deepcopy(QN)
    for qnc in qn_compact:
        indices_compact = states.get_indices_quantumnumbers(qnc, QN_compact)
        QN_compact = states.compact_QN_coupled_indices(QN_compact, indices_compact)
        fields = compact_coupling_field_indices(fields, indices_compact)
    return fields


def compact_coupling_field_indices(
    fields: CouplingFields,
    indices_compact: Union[Sequence[int], npt.NDArray[np.int_]],
) -> CouplingFields:
    field = []
    for cf in fields.fields:
        f = copy.deepcopy(cf.field)
        f = delete_row_column_indices(f, indices_compact)
        field.append(CouplingField(polarization=cf.polarization, field=f))

    return CouplingFields(
        ground_main=fields.ground_main,
        excited_main=fields.excited_main,
        main_coupling=fields.main_coupling,
        ground_states=fields.ground_states,
        excited_states=fields.excited_states,
        fields=field,
    )


def insert_levels_coupling_field(
    fields: CouplingFields,
    indices_insert: Sequence[int],
) -> CouplingFields:
    field = []
    for cf in fields.fields:
        f = copy.deepcopy(cf.field)
        f = insert_row_column_indices(f, indices_insert)
        field.append(CouplingField(polarization=cf.polarization, field=f))
    return CouplingFields(
        ground_main=fields.ground_main,
        excited_main=fields.excited_main,
        main_coupling=fields.main_coupling,
        ground_states=fields.ground_states,
        excited_states=fields.excited_states,
        fields=field,
    )



from typing import List, Sequence, Tuple, overload

import numpy as np
import numpy.typing as npt

from centrex_tlf.states import (
    CoupledBasisState,
    CoupledState,
    UncoupledBasisState,
    UncoupledState,
)

__all__ = ["reorder_evecs", "matrix_to_states", "reduced_basis_hamiltonian"]


def reorder_evecs(
    V_in: npt.NDArray[np.complex128],
    E_in: npt.NDArray[np.complex128],
    V_ref: npt.NDArray[np.complex128],
) -> Tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    """Reshuffle eigenvectors and eigenergies based on a reference

    Args:
        V_in (np.ndarray): eigenvector matrix to be reorganized
        E_in (np.ndarray): energy vector to be reorganized
        V_ref (np.ndarray): reference eigenvector matrix

    Returns:
        (np.ndarray, np.ndarray): energy vector, eigenvector matrix
    """
    # take dot product between each eigenvector in V and state_vec
    overlap_vectors = np.absolute(np.matmul(np.conj(V_in.T), V_ref))

    # find which state has the largest overlap:
    index = np.argsort(np.argmax(overlap_vectors, axis=1))
    # store energy and state
    E_out = E_in[index]
    V_out = V_in[:, index]

    return E_out, V_out


@overload
def matrix_to_states(
    V: npt.NDArray[np.complex128], QN: Sequence[CoupledBasisState]
) -> List[CoupledState]: ...


@overload
def matrix_to_states(
    V: npt.NDArray[np.complex128], QN: Sequence[UncoupledBasisState]
) -> List[UncoupledState]: ...


def matrix_to_states(V, QN):
    """Turn a matrix of eigenvectors into a list of state objects
    QN is in the basis the diagonal Hamiltonian H was formed from corresponding to the
    eigenvectors V.

    Args:
        V (npt.NDArray[np.complex128]): array with columns corresponding to eigenvectors
        QN (Sequence[BasisState]): list of State objects
        E (List, optional): list of energies corresponding to the states.
                            Defaults to None.

    Returns:
        List[State]: list of eigenstates expressed as State objects
    """
    # find dimensions of matrix
    matrix_dimensions = V.shape

    # initialize a list for storing eigenstates
    eigenstates = []

    for i in range(0, matrix_dimensions[1]):
        # find state vector
        state_vector = V[:, i]

        # ensure that largest component has positive sign
        index = np.argmax(np.abs(state_vector))
        state_vector = state_vector * np.sign(state_vector[index])

        data = []

        # get data in correct format for initializing state object
        for j, amp in enumerate(state_vector):
            data.append((amp, QN[j]))

        # store the state in the list
        if isinstance(QN[0], CoupledBasisState):
            state = CoupledState(data)
        elif isinstance(QN[1], UncoupledBasisState):
            state = UncoupledState(data)
        else:
            raise ValueError(
                f"QN should be list of CoupledBasisState or UncoupledBasisState, not {type(QN[0])}"
            )
        eigenstates.append(state)

    # return the list of states
    return eigenstates


def reduced_basis_hamiltonian(
    basis_original: Sequence[CoupledState],
    H_original: npt.NDArray[np.complex128],
    basis_reduced: Sequence[CoupledState],
) -> npt.NDArray[np.complex128]:
    """Generate Hamiltonian for a sub-basis of the original basis

    Args:
        basis_original (Sequence[State],): sequence of states of original basis
        H_original (npt.NDArray[np.complex128]): original Hamiltonian
        basis_reduced (Sequence[State]): sequence of states of sub-basis

    Returns:
        npt.NDArray[np.complex128]: Hamiltonian in sub-basis
    """

    # Determine the indices of each of the reduced basis states
    index_red = np.zeros(len(basis_reduced), dtype=int)
    for i, state_red in enumerate(basis_reduced):
        index_red[i] = basis_original.index(state_red)

    # Initialize matrix for Hamiltonian in reduced basis
    H_red = np.zeros((len(basis_reduced), len(basis_reduced)), dtype=complex)

    # Loop over reduced basis states and pick out the correct matrix elements
    # for the Hamiltonian in the reduced basis
    for i, state_i in enumerate(basis_reduced):
        for j, state_j in enumerate(basis_reduced):
            H_red[i, j] = H_original[index_red[i], index_red[j]]

    return H_red



"""System of equations generation for Lindblad master equation.

This module generates symbolic systems of differential equations from the Lindblad
master equation, which describes the time evolution of the density matrix for an
open quantum system including decoherence and dissipation.

The Lindblad master equation in Lindblad form is:
    dρ/dt = -i[H, ρ] + Σᵢ(CᵢρCᵢ† - ½{Cᵢ†Cᵢ, ρ})

where H is the Hamiltonian, ρ is the density matrix, Cᵢ are Lindblad operators
(jump operators), and {A,B} = AB + BA is the anticommutator.
"""

from __future__ import annotations

from typing import Literal, Tuple, Union, overload

import numpy as np
import numpy.typing as npt
import sympy as smp

from .utils import generate_density_matrix_symbolic

__all__ = ["generate_system_of_equations_symbolic"]


@overload
def generate_system_of_equations_symbolic(
    hamiltonian: smp.Matrix,
    C_array: npt.NDArray[np.floating | np.complexfloating],  # 3D array
    fast: bool,
    split_output: Literal[False],
) -> smp.Matrix: ...


@overload
def generate_system_of_equations_symbolic(
    hamiltonian: smp.Matrix,
    C_array: npt.NDArray[np.floating | np.complexfloating],  # 3D array
    fast: bool,
) -> smp.Matrix: ...


@overload
def generate_system_of_equations_symbolic(
    hamiltonian: smp.Matrix,
    C_array: npt.NDArray[np.floating | np.complexfloating],  # 3D array
    fast: bool,
    split_output: Literal[True],
) -> Tuple[smp.Matrix, smp.Matrix]: ...


def generate_system_of_equations_symbolic(
    hamiltonian: smp.Matrix,
    C_array: npt.NDArray[np.floating | np.complexfloating],  # 3D array
    fast: bool = False,
    split_output: bool = False,
) -> Union[smp.Matrix, Tuple[smp.Matrix, smp.Matrix]]:
    """Generate symbolic system of differential equations from Lindblad master equation.

    Constructs the symbolic representation of the Lindblad master equation:
        dρ/dt = -i[H, ρ] + Σᵢ(CᵢρCᵢ† - ½{Cᵢ†Cᵢ, ρ})

    where H is the Hamiltonian (symbolic matrix), ρ is the density matrix (symbolic),
    and Cᵢ are Lindblad operators (numerical collapse operators).

    This function generates a symbolic matrix equation representing the time evolution
    of each density matrix element ρᵢⱼ. The result can be converted to numerical code
    for efficient ODE solving.

    Args:
        hamiltonian: Symbolic Hamiltonian matrix (n_states × n_states) containing
            sympy symbols for time-dependent parameters (e.g., laser detunings,
            Rabi frequencies). Typically contains Complex symbols for coupling
            strengths and Real symbols for energies.
        C_array: Array of Lindblad/collapse operators with shape (n_operators,
            n_states, n_states). Each C_array[i] represents a decay channel or
            decoherence process. Can be real or complex; will be converted to
            complex128 if real.
        fast: If True, uses sparse matrix multiplication optimization that only
            processes non-zero elements. Significant speedup for sparse collapse
            operators (e.g., spontaneous decay between specific states). Default False.
        split_output: If True, returns Hamiltonian and Lindblad contributions
            separately as a tuple. If False, returns combined system. Default False.

    Returns:
        If split_output=False:
            Symbolic matrix (n_states × n_states) representing dρ/dt with elements
            as symbolic expressions in terms of density matrix elements ρᵢⱼ and
            Hamiltonian parameters.
        If split_output=True:
            Tuple of two matrices:
                - hamiltonian_term: -i[H, ρ] contribution
                - lindblad_term: Combined Lindblad dissipation and decay terms

    Notes:
        - The fast mode assumes sparse C matrices and only processes non-zero entries
        - For dense C matrices or when most elements are non-zero, fast=False is better
        - The symbolic output can be lambdified for numerical integration
        - Typical use: generate symbolic equations once, then solve numerically many times
        - Memory usage scales as O(n_states²) for symbolic expressions

    Raises:
        ValueError: If hamiltonian is not square or C_array dimensions incompatible.

    Example:
        >>> import sympy as smp
        >>> import numpy as np
        >>>
        >>> # Create 2-level system
        >>> n = 2
        >>> Omega = smp.Symbol("Omega", complex=True)  # Rabi frequency
        >>> delta = smp.Symbol("delta", real=True)      # Detuning
        >>>
        >>> # Hamiltonian for driven 2-level system
        >>> H = smp.Matrix([[0, Omega/2], [smp.conjugate(Omega)/2, delta]])
        >>>
        >>> # Spontaneous decay operator |g⟩⟨e|
        >>> Gamma = 1.0  # Decay rate
        >>> C_decay = np.array([[[0, np.sqrt(Gamma)], [0, 0]]])
        >>>
        >>> # Generate equations
        >>> system = generate_system_of_equations_symbolic(H, C_decay, fast=True)
        >>>
        >>> # system now contains symbolic dρ/dt = f(ρ, Omega, delta)
        >>> # Can be lambdified for numerical integration:
        >>> from sympy.utilities.lambdify import lambdify
        >>> rho_symbols = [smp.Symbol(f"rho_{i}_{j}") for i in range(n) for j in range(n)]
        >>> f_numeric = lambdify([rho_symbols, Omega, delta], system, "numpy")
    """
    # Extract system size from Hamiltonian dimensions
    n_states: int = hamiltonian.shape[0]

    # Generate symbolic density matrix with elements ρᵢⱼ as sympy symbols
    density_matrix: smp.Matrix = generate_density_matrix_symbolic(n_states)

    # Ensure collapse operators are complex-valued for proper conjugation
    if not np.iscomplexobj(C_array):
        C_array = C_array.astype(np.complex128)

    # Compute conjugate transpose of all collapse operators: Cᵢ†
    # einsum("ijk->ikj") transposes the last two dimensions for each operator
    C_conj_array: npt.NDArray[np.complexfloating] = np.einsum(
        "ijk->ikj",
        C_array.conj(),  # type: ignore[arg-type]
    )

    # Initialize accumulator for dissipation term: Σᵢ CᵢρCᵢ†
    dissipation_sum: smp.Matrix = smp.zeros(n_states, n_states)

    if fast:
        # Sparse optimization: only compute non-zero contributions
        # Significant speedup when collapse operators are sparse (typical for decay)
        for C, C_conj in zip(C_array, C_conj_array):
            # Get indices of non-zero elements: nonzero returns tuple of (row_indices, col_indices)
            nonzero_C = np.nonzero(C)
            nonzero_C_conj = np.nonzero(C_conj)

            # Only process if both operators have non-zero elements
            if len(nonzero_C[0]) > 0 and len(nonzero_C_conj[0]) > 0:
                # For sparse operators (e.g., |g⟩⟨e|), typically only one non-zero element
                # nonzero_C[0][0] is row index, nonzero_C[1][0] is column index
                # density_matrix indexing returns a MatrixElement directly (no [0] needed)
                value = (
                    C[nonzero_C][0]  # Get first non-zero value from C
                    * C_conj[nonzero_C_conj][0]  # Get first non-zero value from C†
                    * density_matrix[int(nonzero_C[1][0]), int(nonzero_C_conj[0][0])]
                )
                dissipation_sum[int(nonzero_C[0][0]), int(nonzero_C_conj[1][0])] += (
                    value
                )
    else:
        # Standard computation: Σᵢ CᵢρCᵢ† for all operators
        # Use full matrix multiplication (more general but slower)
        for idx in range(C_array.shape[0]):
            dissipation_sum += C_array[idx] @ density_matrix @ C_conj_array[idx]

    # Precompute Σᵢ Cᵢ†Cᵢ for anticommutator term: -½{Cᵢ†Cᵢ, ρ}
    # einsum("ijk,ikl") efficiently computes the sum of Cᵢ†Cᵢ for all i
    C_dagger_C_sum: npt.NDArray[np.complexfloating] = np.einsum(
        "ijk,ikl",
        C_conj_array,  # type: ignore[arg-type]
        C_array,  # type: ignore[arg-type]
    )

    # Anticommutator term: -½{Cᵢ†Cᵢ, ρ} = -½(Cᵢ†Cᵢ·ρ + ρ·Cᵢ†Cᵢ)
    anticommutator_term: smp.Matrix = -0.5 * (
        C_dagger_C_sum @ density_matrix + density_matrix @ C_dagger_C_sum
    )

    # Compute Hamiltonian contribution: -i[H, ρ] = -i(Hρ - ρH)
    # This is the coherent (unitary) evolution part
    hamiltonian_term: smp.Matrix = -1j * (
        hamiltonian @ density_matrix - density_matrix @ hamiltonian
    )

    if split_output:
        # Return coherent and dissipative parts separately
        lindblad_dissipation = dissipation_sum + anticommutator_term
        return hamiltonian_term, lindblad_dissipation
    else:
        # Return complete Lindblad equation: dρ/dt = -i[H,ρ] + Σᵢ(CᵢρCᵢ† - ½{Cᵢ†Cᵢ,ρ})
        system: smp.Matrix = smp.zeros(n_states, n_states)
        system += dissipation_sum + anticommutator_term + hamiltonian_term
        return system



import pickle
import warnings
from pathlib import Path

import numpy as np

from centrex_tlf import hamiltonian, states
from centrex_tlf.hamiltonian.reduced_hamiltonian import (
    ReducedHamiltonian,
    ReducedHamiltonianTotal,
)


def test_generate_reduced_X_hamiltonian():
    X_states_approx = states.generate_coupled_states_ground(Js=[0, 1, 2, 3])

    H_reduced = hamiltonian.reduced_hamiltonian.generate_reduced_X_hamiltonian(
        X_states_approx=X_states_approx,
        Jmin=0,
        Jmax=3,
    )

    with open(Path(__file__).parent / "X_reduced.pkl", "rb") as f:
        H_reduced_test: hamiltonian.reduced_hamiltonian.ReducedHamiltonian = (
            pickle.load(f)
        )

    assert np.allclose(H_reduced.H, H_reduced_test.H, rtol=1e-15, atol=1e-3)
    # assert np.allclose(H_reduced.V, H_reduced_test.V)
    assert len(H_reduced.QN_basis) == len(H_reduced_test.QN_basis)
    assert len(H_reduced.QN_construct) == len(H_reduced_test.QN_construct)
    assert H_reduced.QN_construct == H_reduced_test.QN_construct

    state_vectors = np.array(
        [s.state_vector(H_reduced.QN_basis) for s in H_reduced_test.QN_basis]
    )

    assert np.allclose(np.trace(state_vectors), len(state_vectors), rtol=1e-8)


def test_generate_reduced_B_hamiltonian_omega():
    # omega basis
    B_states_approx = states.generate_coupled_states_excited(
        Js=[1, 2, 3], Ps=None, Omegas=[-1, 1]
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        H_reduced_omega = (
            hamiltonian.reduced_hamiltonian.generate_reduced_B_hamiltonian(
                B_states_approx=B_states_approx, Jmin=1, Jmax=5, rtol=1e-5
            )
        )

    with open(Path(__file__).parent / "B_reduced_omega.pkl", "rb") as f:
        H_reduced_test_omega: hamiltonian.reduced_hamiltonian.ReducedHamiltonian = (
            pickle.load(f)
        )

    assert np.allclose(H_reduced_omega.H, H_reduced_test_omega.H, rtol=1e-15, atol=1e-3)
    # assert np.allclose(H_reduced_omega.V, H_reduced_test_omega.V)
    assert len(H_reduced_omega.QN_basis) == len(H_reduced_test_omega.QN_basis)
    assert len(H_reduced_omega.QN_construct) == len(H_reduced_test_omega.QN_construct)
    assert H_reduced_omega.QN_construct == H_reduced_test_omega.QN_construct

    state_vectors = np.array(
        [
            s.state_vector(H_reduced_omega.QN_basis)
            for s in H_reduced_test_omega.QN_basis
        ]
    )

    assert np.allclose(np.trace(state_vectors), len(state_vectors), rtol=1e-6)


def test_generate_reduced_B_hamiltonian_parity():
    # parity basis
    B_states_approx = states.generate_coupled_states_excited(
        Js=[1, 2, 3], Ps=[-1, 1], Omegas=1
    )

    H_reduced_parity = hamiltonian.reduced_hamiltonian.generate_reduced_B_hamiltonian(
        B_states_approx=B_states_approx, Jmin=1, Jmax=5, rtol=1e-5
    )

    with open(Path(__file__).parent / "B_reduced_parity.pkl", "rb") as f:
        H_reduced_parity_test: ReducedHamiltonian = pickle.load(f)

    assert np.allclose(
        H_reduced_parity.H, H_reduced_parity_test.H, rtol=1e-15, atol=1e-3
    )
    # assert np.allclose(H_reduced_parity.V, H_reduced_parity_test.V)
    assert len(H_reduced_parity.QN_basis) == len(H_reduced_parity_test.QN_basis)
    assert len(H_reduced_parity.QN_construct) == len(H_reduced_parity_test.QN_construct)
    assert H_reduced_parity.QN_construct == H_reduced_parity_test.QN_construct

    state_vectors = np.array(
        [
            s.state_vector(H_reduced_parity.QN_basis)
            for s in H_reduced_parity_test.QN_basis
        ]
    )

    assert np.allclose(np.trace(state_vectors), len(state_vectors), rtol=1e-6)


def test_generate_total_reduced_hamiltonian():
    B_states_approx = states.generate_coupled_states_excited(
        Js=[1, 2, 3], Ps=[-1, 1], Omegas=1
    )
    X_states_approx = states.generate_coupled_states_ground(Js=[0, 1, 2, 3])

    H_reduced_total = (
        hamiltonian.reduced_hamiltonian.generate_total_reduced_hamiltonian(
            X_states_approx=X_states_approx,
            B_states_approx=B_states_approx,
            rtol=None,
            stol=1e-3,
            Jmin_X=0,
            Jmax_X=3,
            Jmin_B=1,
            Jmax_B=5,
        )
    )

    with open(Path(__file__).parent / "H_reduced_total.pkl", "rb") as f:
        H_reduced_total_test: ReducedHamiltonianTotal = pickle.load(f)

    assert np.allclose(
        H_reduced_total.H_int, H_reduced_total_test.H_int, rtol=1e-15, atol=1e-3
    )
    # assert np.allclose(H_reduced_total.V_ref_int, H_reduced_total_test.V_ref_int)
    assert len(H_reduced_total.QN) == len(H_reduced_total_test.QN)
    assert len(H_reduced_total.QN_basis) == len(H_reduced_total_test.QN_basis)
    assert H_reduced_total.QN_basis == H_reduced_total_test.QN_basis

    state_vectors = np.array(
        [s.state_vector(H_reduced_total.QN) for s in H_reduced_total_test.QN]
    )

    assert np.allclose(np.trace(state_vectors), len(state_vectors), rtol=1e-6)



import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

from centrex_tlf import hamiltonian
from centrex_tlf.constants import TlFNuclearSpins
from centrex_tlf.hamiltonian.basis_transformations import generate_transform_matrix
from centrex_tlf.states.find_states import QuantumSelector
from centrex_tlf.states.generate_states import (
    generate_coupled_states_ground,
    generate_coupled_states_X,
    generate_uncoupled_states_ground,
)

# generate the hyperfine sublevels in J=2
qn_select = QuantumSelector(J=[2])
QN = generate_coupled_states_X(qn_select)

# generate the transformation matrix
qn = generate_uncoupled_states_ground(
    Js=np.arange(0, 6 + 1), nuclear_spins=TlFNuclearSpins()
)
qnc = generate_coupled_states_ground(
    Js=np.arange(0, 6 + 1), nuclear_spins=TlFNuclearSpins()
)
transform = generate_transform_matrix(qn, qnc)

# generate the X hamiltonian terms
H = hamiltonian.generate_uncoupled_hamiltonian_X(qn)

# create a function outputting the hamiltonian as a function of E and B
Hfunc = hamiltonian.generate_uncoupled_hamiltonian_X_function(H)

# generate the reduced X hamiltonian
QN_states, H = hamiltonian.generate_reduced_X_hamiltonian(
    QN, Jmin=0, Jmax=6, transform=transform, H_func=Hfunc
)

E, V = np.linalg.eigh(H)

# V/cm
Ez = np.linspace(0, 15e3, 51)

# original eigenvectors used in tracking states as energies change order
V_track = V.copy()

# empty array for storing energies
energy = np.zeros([Ez.size, len(QN)], dtype=np.complex128)

# iterate over the electric field values
for idx, Ei in tqdm(enumerate(Ez), total=len(Ez)):
    _, H = hamiltonian.generate_reduced_X_hamiltonian(
        QN,
        E=[0.0, 0.0, Ei],
        B=[0.0, 0.0, 1e-3],
        Jmin=0,
        Jmax=6,
        transform=transform,
        H_func=Hfunc,
    )
    E, V = np.linalg.eigh(H)

    # sort indices to keep the state order the same
    indices = np.argmax(np.abs(V_track.conj().T @ V), axis=1)
    energy[idx, :] = E[indices]
    V_track[:, :] = V[:, indices]

# plot the J=2, mJ=0 Stark curves
fig, ax = plt.subplots(figsize=(12, 8))
ax.plot(Ez, (energy.real - energy.real[0, 0]) / (2 * np.pi * 1e9), lw=3, color="k")
ax.set_xlabel("E [V/cm]")
ax.set_ylabel("Energy [GHz]")
ax.set_title("|J=2> Stark Curve")
ax.grid(True)

plt.show()



from typing import List, Sequence

import numpy as np
import numpy.typing as npt

from .states import CoupledState

__all__ = ["compact_QN_coupled_indices"]


def compact_QN_coupled_indices(
    QN: Sequence[CoupledState], indices_compact: npt.NDArray[np.int_]
) -> List[CoupledState]:
    """Compact the states given by indices in indices_compact

    Args:
        QN (list): states
        indices_compact (list, array): indices to compact into a single state

    Returns:
        list: compacted states
    """
    QNc = [QN[idx] for idx in indices_compact]

    def slc(s):
        return s.largest

    Js = np.unique([slc(s).J for s in QNc if slc(s).J is not None])
    F1s = np.unique([slc(s).F1 for s in QNc if slc(s).F1 is not None])
    Fs = np.unique([slc(s).F for s in QNc if slc(s).F is not None])
    mFs = np.unique([slc(s).mF for s in QNc if slc(s).mF is not None])
    Ps = np.unique([slc(s).P for s in QNc if slc(s).P is not None])

    QNcompact = [qn for idx, qn in enumerate(QN) if idx not in indices_compact[1:]]

    state_rep = QNcompact[indices_compact[0]].largest
    if len(Js) != 1:
        state_rep.J = None
    if len(F1s) != 1:
        state_rep.F1 = None
    if len(Fs) != 1:
        state_rep.F = None
    if len(mFs) != 1:
        state_rep.mF = None
    if len(Ps) != 1:
        state_rep.P = None

    # make it a state again instead of uncoupled basisstate
    QNcompact[indices_compact[0]] = (1.0 + 0j) * state_rep

    return QNcompact
